#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: September 24, 2020
#
"""It is utilities for plotting figures."""

import os
import numpy as np
try:
    import matplotlib.pyplot as plt
except:
    plt = None
try:
    import pylab
except:
    pylab = None

from bayeso.utils import utils_common
from bayeso.utils import utils_logger
from bayeso import constants

logger = utils_logger.get_logger('utils_plotting')


@utils_common.validate_types
def _set_font_config(use_tex: bool) -> constants.TYPE_NONE: # pragma: no cover
    """
    It sets a font configuration.

    :param use_tex: flag for using latex.
    :type use_tex: bool.

    :returns: None.
    :rtype: NoneType

    """

    if use_tex:
        plt.rc('text', usetex=True)
    else:
        plt.rc('pdf', fonttype=42)

@utils_common.validate_types
def _set_ax_config(ax: 'matplotlib.axes._subplots.AxesSubplot', str_x_axis: str, str_y_axis: str,
    size_labels: int=32,
    size_ticks: int=22,
    xlim_min: constants.TYPING_UNION_FLOAT_NONE=None,
    xlim_max: constants.TYPING_UNION_FLOAT_NONE=None,
    draw_box: bool=True,
    draw_zero_axis: bool=False,
    draw_grid: bool=True,
) -> constants.TYPE_NONE: # pragma: no cover
    """
    It sets an axis configuration.

    :param ax: inputs for acquisition function. Shape: (n, d).
    :type ax: matplotlib.axes._subplots.AxesSubplot
    :param str_x_axis: the name of x axis.
    :type str_x_axis: str.
    :param str_y_axis: the name of y axis.
    :type str_y_axis: str.
    :param size_labels: label size.
    :type size_labels: int., optional
    :param size_ticks: tick size.
    :type size_ticks: int., optional
    :param xlim_min: None, or minimum for x limit.
    :type xlim_min: NoneType or float, optional
    :param xlim_max: None, or maximum for x limit.
    :type xlim_max: NoneType or float, optional
    :param draw_box: flag for drawing a box.
    :type draw_box: bool., optional
    :param draw_zero_axis: flag for drawing a zero axis.
    :type draw_zero_axis: bool., optional
    :param draw_grid: flag for drawing grids.
    :type draw_grid: bool., optional

    :returns: None.
    :rtype: NoneType

    """

    if str_x_axis is not None:
        ax.set_xlabel(str_x_axis, fontsize=size_labels)
    ax.set_ylabel(str_y_axis, fontsize=size_labels)
    ax.tick_params(labelsize=size_ticks)

    if not draw_box:
        ax.spines['top'].set_color('none')
        ax.spines['right'].set_color('none')
    if xlim_min is not None and xlim_max is not None:
        ax.set_xlim([xlim_min, xlim_max])
    if draw_zero_axis:
        ax.spines['bottom'].set_position('zero')
    if draw_grid:
        ax.grid()

@utils_common.validate_types
def _save_figure(path_save: str, str_postfix: str,
    str_prefix: str=''
) -> constants.TYPE_NONE: # pragma: no cover
    """
    It saves a figure.

    :param path_save: path for saving a figure.
    :type path_save: str.
    :param str_postfix: the name of postfix.
    :type str_postfix: str.
    :param str_prefix: the name of prefix.
    :type str_prefix: str., optional

    :returns: None.
    :rtype: NoneType

    """

    if path_save is not None and str_postfix is not None:
        str_figure = str_prefix + str_postfix
        plt.savefig(os.path.join(path_save, str_figure + '.pdf'),
            format='pdf', transparent=True, bbox_inches='tight')

@utils_common.validate_types
def _show_figure(pause_figure: bool, time_pause: constants.TYPING_UNION_INT_FLOAT
) -> constants.TYPE_NONE: # pragma: no cover
    """
    It shows a figure.

    :param pause_figure: flag for pausing before closing a figure.
    :type pause_figure: bool.
    :param time_pause: pausing time.
    :type time_pause: int. or float

    :returns: None.
    :rtype: NoneType

    """

    if pause_figure:
        if time_pause < np.inf:
            plt.ion()
            plt.pause(time_pause)
            plt.close('all')
        else:
            plt.show()

@utils_common.validate_types
def plot_gp_via_sample(X: np.ndarray, Ys: np.ndarray,
    path_save: constants.TYPING_UNION_STR_NONE=None,
    str_postfix: constants.TYPING_UNION_STR_NONE=None,
    str_x_axis: str='x',
    str_y_axis: str='y',
    use_tex: bool=False,
    draw_zero_axis: bool=False,
    pause_figure: bool=True,
    time_pause: constants.TYPING_UNION_INT_FLOAT=constants.TIME_PAUSE,
    colors: np.ndarray=constants.COLORS,
) -> constants.TYPE_NONE: # pragma: no cover
    """
    It is for plotting sampled functions from multivariate distributions.

    :param X: training inputs. Shape: (n, 1).
    :type X: numpy.ndarray
    :param Ys: training outputs. Shape: (m, n).
    :type Ys: numpy.ndarray
    :param path_save: None, or path for saving a figure.
    :type path_save: NoneType or str., optional
    :param str_postfix: None, or the name of postfix.
    :type str_postfix: NoneType or str., optional
    :param str_x_axis: the name of x axis.
    :type str_x_axis: str., optional
    :param str_y_axis: the name of y axis.
    :type str_y_axis: str., optional
    :param use_tex: flag for using latex.
    :type use_tex: bool., optional
    :param draw_zero_axis: flag for drawing a zero axis.
    :type draw_zero_axis: bool., optional
    :param pause_figure: flag for pausing before closing a figure.
    :type pause_figure: bool., optional
    :param time_pause: pausing time.
    :type time_pause: int. or float, optional
    :param colors: array of colors.
    :type colors: np.ndarray, optional

    :returns: None.
    :rtype: NoneType

    :raises: AssertionError

    """

    assert isinstance(X, np.ndarray)
    assert isinstance(Ys, np.ndarray)
    assert isinstance(path_save, (str, type(None)))
    assert isinstance(str_postfix, (str, type(None)))
    assert isinstance(str_x_axis, str)
    assert isinstance(str_y_axis, str)
    assert isinstance(use_tex, bool)
    assert isinstance(draw_zero_axis, bool)
    assert isinstance(pause_figure, bool)
    assert isinstance(time_pause, (int, float))
    assert isinstance(colors, np.ndarray)
    assert len(X.shape) == 2
    assert len(Ys.shape) == 2
    assert X.shape[1] == 1
    assert X.shape[0] == Ys.shape[1]

    if plt is None or pylab is None:
        logger.info('matplotlib or pylab is not installed.')
        return
    _set_font_config(use_tex)

    _ = plt.figure(figsize=(8, 6))
    ax = plt.gca()

    for Y in Ys:
        ax.plot(X.flatten(), Y,
            c=colors[0],
            lw=4,
            alpha=0.3,
        )

    _set_ax_config(ax, str_x_axis, str_y_axis, xlim_min=np.min(X), xlim_max=np.max(X),
        draw_zero_axis=draw_zero_axis)

    plt.tight_layout()

    if path_save is not None and str_postfix is not None:
        _save_figure(path_save, str_postfix, str_prefix='gp_sampled_')
    _show_figure(pause_figure, time_pause)

@utils_common.validate_types
def plot_gp_via_distribution(X_train: np.ndarray, Y_train: np.ndarray,
    X_test: np.ndarray, mean_test: np.ndarray, std_test: np.ndarray,
    Y_test: constants.TYPING_UNION_ARRAY_NONE=None,
    path_save: constants.TYPING_UNION_STR_NONE=None,
    str_postfix: constants.TYPING_UNION_STR_NONE=None,
    str_x_axis: str='x',
    str_y_axis: str='y',
    use_tex: bool=False,
    draw_zero_axis: bool=False,
    pause_figure: bool=True,
    time_pause: constants.TYPING_UNION_INT_FLOAT=constants.TIME_PAUSE,
    range_shade: float=constants.RANGE_SHADE,
    colors: np.ndarray=constants.COLORS,
) -> constants.TYPE_NONE: # pragma: no cover
    """
    It is for plotting Gaussian process regression.

    :param X_train: training inputs. Shape: (n, 1).
    :type X_train: numpy.ndarray
    :param Y_train: training outputs. Shape: (n, 1).
    :type Y_train: numpy.ndarray
    :param X_test: test inputs. Shape: (m, 1).
    :type X_test: numpy.ndarray
    :param mean_test: posterior predictive mean function values over `X_test`.
        Shape: (m, 1).
    :type mean_test: numpy.ndarray
    :param std_test: posterior predictive standard deviation function values
        over `X_test`. Shape: (m, 1).
    :type std_test: numpy.ndarray
    :param Y_test: None, or true test outputs. Shape: (m, 1).
    :type Y_test: NoneType or numpy.ndarray, optional
    :param path_save: None, or path for saving a figure.
    :type path_save: NoneType or str., optional
    :param str_postfix: None, or the name of postfix.
    :type str_postfix: NoneType or str., optional
    :param str_x_axis: the name of x axis.
    :type str_x_axis: str., optional
    :param str_y_axis: the name of y axis.
    :type str_y_axis: str., optional
    :param use_tex: flag for using latex.
    :type use_tex: bool., optional
    :param draw_zero_axis: flag for drawing a zero axis.
    :type draw_zero_axis: bool., optional
    :param pause_figure: flag for pausing before closing a figure.
    :type pause_figure: bool., optional
    :param time_pause: pausing time.
    :type time_pause: int. or float, optional
    :param range_shade: shade range for standard deviation.
    :type range_shade: float, optional
    :param colors: array of colors.
    :type colors: np.ndarray, optional

    :returns: None.
    :rtype: NoneType

    :raises: AssertionError

    """

    assert isinstance(X_train, np.ndarray)
    assert isinstance(Y_train, np.ndarray)
    assert isinstance(X_test, np.ndarray)
    assert isinstance(mean_test, np.ndarray)
    assert isinstance(std_test, np.ndarray)
    assert isinstance(Y_test, (np.ndarray, type(None)))
    assert isinstance(path_save, (str, type(None)))
    assert isinstance(str_postfix, (str, type(None)))
    assert isinstance(str_x_axis, str)
    assert isinstance(str_y_axis, str)
    assert isinstance(use_tex, bool)
    assert isinstance(draw_zero_axis, bool)
    assert isinstance(pause_figure, bool)
    assert isinstance(time_pause, (int, float))
    assert isinstance(range_shade, float)
    assert isinstance(colors, np.ndarray)
    assert len(X_train.shape) == 2
    assert len(X_test.shape) == 2
    assert len(Y_train.shape) == 2
    assert len(mean_test.shape) == 2
    assert len(std_test.shape) == 2
    assert X_train.shape[1] == X_test.shape[1] == 1
    assert Y_train.shape[1] == 1
    assert X_train.shape[0] == Y_train.shape[0]
    assert mean_test.shape[1] == 1
    assert std_test.shape[1] == 1
    assert X_test.shape[0] == mean_test.shape[0] == std_test.shape[0]
    if Y_test is not None:
        assert len(Y_test.shape) == 2
        assert Y_test.shape[1] == 1
        assert X_test.shape[0] == Y_test.shape[0]

    if plt is None or pylab is None:
        logger.info('matplotlib or pylab is not installed.')
        return
    _set_font_config(use_tex)

    _ = plt.figure(figsize=(8, 6))
    ax = plt.gca()

    if Y_test is not None:
        ax.plot(X_test.flatten(), Y_test.flatten(),
            c=colors[1],
            linewidth=4,
            marker='None')
    ax.plot(X_test.flatten(), mean_test.flatten(),
        c=colors[2],
        linewidth=4,
        marker='None')
    ax.fill_between(X_test.flatten(),
        mean_test.flatten() - range_shade * std_test.flatten(),
        mean_test.flatten() + range_shade * std_test.flatten(),
        color=colors[2],
        alpha=0.3)
    ax.plot(X_train.flatten(), Y_train.flatten(),
        'x',
        c=colors[0],
        markersize=10,
        mew=4)

    _set_ax_config(ax, str_x_axis, str_y_axis, xlim_min=np.min(X_test),
        xlim_max=np.max(X_test), draw_zero_axis=draw_zero_axis)

    plt.tight_layout()

    if path_save is not None and str_postfix is not None:
        _save_figure(path_save, str_postfix, str_prefix='gp_')
    _show_figure(pause_figure, time_pause)

@utils_common.validate_types
def plot_minimum_vs_iter(minima: np.ndarray, list_str_label: constants.TYPING_LIST[str],
    num_init: int, draw_std: bool,
    include_marker: bool=True,
    include_legend: bool=False,
    use_tex: bool=False,
    path_save: constants.TYPING_UNION_STR_NONE=None,
    str_postfix: constants.TYPING_UNION_STR_NONE=None,
    str_x_axis: str='Iteration',
    str_y_axis: str='Minimum function value',
    pause_figure: bool=True,
    time_pause: constants.TYPING_UNION_INT_FLOAT=constants.TIME_PAUSE,
    range_shade: float=constants.RANGE_SHADE,
    markers: np.ndarray=constants.MARKERS,
    colors: np.ndarray=constants.COLORS,
) -> constants.TYPE_NONE: # pragma: no cover
    """
    It is for plotting optimization results of Bayesian optimization, in
    terms of iterations.

    :param minima: function values over acquired examples. Shape: (b, r, n)
        where b is the number of experiments, r is the number of rounds,
        and n is the number of iterations per round.
    :type minima: numpy.ndarray
    :param list_str_label: list of label strings. Shape: (b, ).
    :type list_str_label: list
    :param num_init: the number of initial examples < n.
    :type num_init: int.
    :param draw_std: flag for drawing standard deviations.
    :type draw_std: bool.
    :param include_marker: flag for drawing markers.
    :type include_marker: bool., optional
    :param include_legend: flag for drawing a legend.
    :type include_legend: bool., optional
    :param use_tex: flag for using latex.
    :type use_tex: bool., optional
    :param path_save: None, or path for saving a figure.
    :type path_save: NoneType or str., optional
    :param str_postfix: None, or the name of postfix.
    :type str_postfix: NoneType or str., optional
    :param str_x_axis: the name of x axis.
    :type str_x_axis: str., optional
    :param str_y_axis: the name of y axis.
    :type str_y_axis: str., optional
    :param pause_figure: flag for pausing before closing a figure.
    :type pause_figure: bool., optional
    :param time_pause: pausing time.
    :type time_pause: int. or float, optional
    :param range_shade: shade range for standard deviation.
    :type range_shade: float, optional
    :param markers: array of markers.
    :type markers: np.ndarray, optional
    :param colors: array of colors.
    :type colors: np.ndarray, optional

    :returns: None.
    :rtype: NoneType

    :raises: AssertionError

    """

    assert isinstance(minima, np.ndarray)
    assert isinstance(list_str_label, list)
    assert isinstance(num_init, int)
    assert isinstance(draw_std, bool)
    assert isinstance(include_marker, bool)
    assert isinstance(include_legend, bool)
    assert isinstance(use_tex, bool)
    assert isinstance(path_save, (str, type(None)))
    assert isinstance(str_postfix, (str, type(None)))
    assert isinstance(str_x_axis, str)
    assert isinstance(str_y_axis, str)
    assert isinstance(pause_figure, bool)
    assert isinstance(time_pause, (int, float))
    assert isinstance(range_shade, float)
    assert isinstance(markers, np.ndarray)
    assert isinstance(colors, np.ndarray)
    assert len(minima.shape) == 3
    assert minima.shape[0] == len(list_str_label)
    assert minima.shape[2] >= num_init

    if plt is None or pylab is None:
        logger.info('matplotlib or pylab is not installed.')
        return
    _set_font_config(use_tex)

    _ = plt.figure(figsize=(8, 6))
    ax = plt.gca()

    for ind_minimum, arr_minimum in enumerate(minima):
        ind_color = ind_minimum % len(colors)
        ind_marker = ind_minimum % len(markers)

        _, mean_min, std_min = utils_common.get_minimum(arr_minimum, num_init)
        x_data = range(0, mean_min.shape[0])
        y_data = mean_min
        std_data = std_min

        if include_marker:
            ax.plot(x_data, y_data,
                label=list_str_label[ind_minimum],
                c=colors[ind_color],
                linewidth=4,
                marker=markers[ind_marker],
                markersize=10,
                mew=3,
            )
        else:
            ax.plot(x_data, y_data,
                label=list_str_label[ind_minimum],
                c=colors[ind_color],
                linewidth=4,
                marker='None')

        if draw_std:
            ax.fill_between(x_data,
                y_data - range_shade * std_data,
                y_data + range_shade * std_data,
                color=colors[ind_color],
                alpha=0.3)
    lines, _ = ax.get_legend_handles_labels()

    _set_ax_config(ax, str_x_axis, str_y_axis, xlim_min=0, xlim_max=mean_min.shape[0]-1)

    if include_legend:
        plt.legend(loc='upper right', fancybox=False, edgecolor='black', fontsize=24)

    plt.tight_layout()

    if path_save is not None and str_postfix is not None:
        if draw_std:
            str_figure = 'minimum_mean_std_' + str_postfix
        else:
            str_figure = 'minimum_mean_only_' + str_postfix
        _save_figure(path_save, str_figure)

        fig_legend = pylab.figure(figsize=(3, 2))
        fig_legend.legend(lines, list_str_label, 'center', fancybox=False,
            edgecolor='black', fontsize=32)
        fig_legend.savefig(os.path.join(path_save, 'legend_{}.pdf'.format(str_postfix)),
            format='pdf', transparent=True, bbox_inches='tight')

    _show_figure(pause_figure, time_pause)

@utils_common.validate_types
def plot_minimum_vs_time(times: np.ndarray, minima: np.ndarray,
    list_str_label: constants.TYPING_LIST[str], num_init: int, draw_std: bool,
    include_marker: bool=True,
    include_legend: bool=False,
    use_tex: bool=False,
    path_save: constants.TYPING_UNION_STR_NONE=None,
    str_postfix: constants.TYPING_UNION_STR_NONE=None,
    str_x_axis: str='Time (sec.)',
    str_y_axis: str='Minimum function value',
    pause_figure: bool=True,
    time_pause: constants.TYPING_UNION_INT_FLOAT=constants.TIME_PAUSE,
    range_shade: float=constants.RANGE_SHADE,
    markers: np.ndarray=constants.MARKERS,
    colors: np.ndarray=constants.COLORS,
) -> constants.TYPE_NONE: # pragma: no cover
    """
    It is for plotting optimization results of Bayesian optimization, in terms of execution time.

    :param times: execution times. Shape: (b, r, n), or (b, r, `num_init` + n)
        where b is the number of experiments, r is the number of rounds,
        and n is the number of iterations per round.
    :type times: numpy.ndarray
    :param minima: function values over acquired examples. Shape: (b, r, `num_init` + n)
        where b is the number of experiments, r is the number of rounds,
        and n is the number of iterations per round.
    :type minima: numpy.ndarray
    :param list_str_label: list of label strings. Shape: (b, ).
    :type list_str_label: list
    :param num_init: the number of initial examples.
    :type num_init: int.
    :param draw_std: flag for drawing standard deviations.
    :type draw_std: bool.
    :param include_marker: flag for drawing markers.
    :type include_marker: bool., optional
    :param include_legend: flag for drawing a legend.
    :type include_legend: bool., optional
    :param use_tex: flag for using latex.
    :type use_tex: bool., optional
    :param path_save: None, or path for saving a figure.
    :type path_save: NoneType or str., optional
    :param str_postfix: None, or the name of postfix.
    :type str_postfix: NoneType or str., optional
    :param str_x_axis: the name of x axis.
    :type str_x_axis: str., optional
    :param str_y_axis: the name of y axis.
    :type str_y_axis: str., optional
    :param pause_figure: flag for pausing before closing a figure.
    :type pause_figure: bool., optional
    :param time_pause: pausing time.
    :type time_pause: int. or float, optional
    :param range_shade: shade range for standard deviation.
    :type range_shade: float, optional
    :param markers: array of markers.
    :type markers: np.ndarray, optional
    :param colors: array of colors.
    :type colors: np.ndarray, optional

    :returns: None.
    :rtype: NoneType

    :raises: AssertionError

    """

    assert isinstance(times, np.ndarray)
    assert isinstance(minima, np.ndarray)
    assert isinstance(list_str_label, list)
    assert isinstance(num_init, int)
    assert isinstance(draw_std, bool)
    assert isinstance(include_marker, bool)
    assert isinstance(include_legend, bool)
    assert isinstance(use_tex, bool)
    assert isinstance(path_save, (str, type(None)))
    assert isinstance(str_postfix, (str, type(None)))
    assert isinstance(str_x_axis, str)
    assert isinstance(str_y_axis, str)
    assert isinstance(pause_figure, bool)
    assert isinstance(time_pause, (int, float))
    assert isinstance(range_shade, float)
    assert isinstance(markers, np.ndarray)
    assert isinstance(colors, np.ndarray)
    assert len(times.shape) == 3
    assert len(minima.shape) == 3
    assert times.shape[0] == minima.shape[0] == len(list_str_label)
    assert times.shape[1] == minima.shape[1]
    assert minima.shape[2] >= num_init
    assert times.shape[2] == minima.shape[2] or times.shape[2] + num_init == minima.shape[2]

    if plt is None or pylab is None:
        logger.info('matplotlib or pylab is not installed.')
        return
    _set_font_config(use_tex)

    _ = plt.figure(figsize=(8, 6))
    ax = plt.gca()

    list_x_data = []
    for ind_minimum, arr_minimum in enumerate(minima):
        ind_color = ind_minimum % len(colors)
        ind_marker = ind_minimum % len(markers)

        _, mean_min, std_min = utils_common.get_minimum(arr_minimum, num_init)
        x_data = utils_common.get_time(times[ind_minimum], num_init,
            times.shape[2] == minima.shape[2])
        y_data = mean_min
        std_data = std_min

        if include_marker:
            ax.plot(x_data, y_data,
                label=list_str_label[ind_minimum],
                c=colors[ind_color],
                linewidth=4,
                marker=markers[ind_marker],
                markersize=10,
                mew=3,
            )
        else:
            ax.plot(x_data, y_data,
                label=list_str_label[ind_minimum],
                c=colors[ind_color],
                linewidth=4,
                marker='None')

        if draw_std:
            ax.fill_between(x_data,
                y_data - range_shade * std_data,
                y_data + range_shade * std_data,
                color=colors[ind_color],
                alpha=0.3)
        list_x_data.append(x_data)
    lines, _ = ax.get_legend_handles_labels()

    _set_ax_config(ax, str_x_axis, str_y_axis, xlim_min=np.min(list_x_data),
        xlim_max=np.max(list_x_data))

    if include_legend:
        plt.legend(loc='upper right', fancybox=False, edgecolor='black', fontsize=24)

    plt.tight_layout()

    if path_save is not None and str_postfix is not None:
        if draw_std:
            str_figure = 'minimum_time_mean_std_' + str_postfix
        else:
            str_figure = 'minimum_time_mean_only_' + str_postfix
        _save_figure(path_save, str_figure)

        fig_legend = pylab.figure(figsize=(3, 2))
        fig_legend.legend(lines, list_str_label, 'center', fancybox=False,
            edgecolor='black', fontsize=32)
        fig_legend.savefig(os.path.join(path_save, 'legend_{}.pdf'.format(str_postfix)),
            format='pdf', transparent=True, bbox_inches='tight')

    _show_figure(pause_figure, time_pause)

@utils_common.validate_types
def plot_bo_step(X_train: np.ndarray, Y_train: np.ndarray,
    X_test: np.ndarray, Y_test: np.ndarray,
    mean_test: np.ndarray, std_test: np.ndarray,
    path_save: constants.TYPING_UNION_STR_NONE=None,
    str_postfix: constants.TYPING_UNION_STR_NONE=None,
    str_x_axis: str='x',
    str_y_axis: str='y',
    num_init: constants.TYPING_UNION_INT_NONE=None,
    use_tex: bool=False,
    draw_zero_axis: bool=False,
    pause_figure: bool=True,
    time_pause: constants.TYPING_UNION_INT_FLOAT=constants.TIME_PAUSE,
    range_shade: float=constants.RANGE_SHADE,
) -> constants.TYPE_NONE: # pragma: no cover
    """
    It is for plotting Bayesian optimization results step by step.

    :param X_train: training inputs. Shape: (n, 1).
    :type X_train: numpy.ndarray
    :param Y_train: training outputs. Shape: (n, 1).
    :type Y_train: numpy.ndarray
    :param X_test: test inputs. Shape: (m, 1).
    :type X_test: numpy.ndarray
    :param Y_test: true test outputs. Shape: (m, 1).
    :type Y_test: numpy.ndarray
    :param mean_test: posterior predictive mean function values over `X_test`.
        Shape: (m, 1).
    :type mean_test: numpy.ndarray
    :param std_test: posterior predictive standard deviation function values
        over `X_test`. Shape: (m, 1).
    :type std_test: numpy.ndarray
    :param path_save: None, or path for saving a figure.
    :type path_save: NoneType or str., optional
    :param str_postfix: None, or the name of postfix.
    :type str_postfix: NoneType or str., optional
    :param str_x_axis: the name of x axis.
    :type str_x_axis: str., optional
    :param str_y_axis: the name of y axis.
    :type str_y_axis: str., optional
    :param num_init: None, or the number of initial examples.
    :type num_init: NoneType or int., optional
    :param use_tex: flag for using latex.
    :type use_tex: bool., optional
    :param draw_zero_axis: flag for drawing a zero axis.
    :type draw_zero_axis: bool., optional
    :param pause_figure: flag for pausing before closing a figure.
    :type pause_figure: bool., optional
    :param time_pause: pausing time.
    :type time_pause: int. or float, optional
    :param range_shade: shade range for standard deviation.
    :type range_shade: float, optional

    :returns: None.
    :rtype: NoneType

    :raises: AssertionError

    """

    assert isinstance(X_train, np.ndarray)
    assert isinstance(Y_train, np.ndarray)
    assert isinstance(X_test, np.ndarray)
    assert isinstance(Y_test, np.ndarray)
    assert isinstance(mean_test, np.ndarray)
    assert isinstance(std_test, np.ndarray)
    assert isinstance(path_save, (str, type(None)))
    assert isinstance(str_postfix, (str, type(None)))
    assert isinstance(str_x_axis, str)
    assert isinstance(str_y_axis, str)
    assert isinstance(num_init, (int, type(None)))
    assert isinstance(use_tex, bool)
    assert isinstance(draw_zero_axis, bool)
    assert isinstance(pause_figure, bool)
    assert isinstance(time_pause, (int, float))
    assert isinstance(range_shade, float)
    assert len(X_train.shape) == 2
    assert len(X_test.shape) == 2
    assert len(Y_train.shape) == 2
    assert len(Y_test.shape) == 2
    assert len(mean_test.shape) == 2
    assert len(std_test.shape) == 2
    assert X_train.shape[1] == X_test.shape[1] == 1
    assert Y_train.shape[1] == Y_test.shape[1] == 1
    assert X_train.shape[0] == Y_train.shape[0]
    assert mean_test.shape[1] == 1
    assert std_test.shape[1] == 1
    assert X_test.shape[0] == Y_test.shape[0] == mean_test.shape[0] == std_test.shape[0]
    if num_init is not None:
        assert X_train.shape[0] >= num_init

    if plt is None or pylab is None:
        logger.info('matplotlib or pylab is not installed.')
        return
    _set_font_config(use_tex)

    _ = plt.figure(figsize=(8, 6))
    ax = plt.gca()

    ax.plot(X_test, Y_test, 'g', linewidth=4)
    ax.plot(X_test, mean_test, 'b', linewidth=4)
    ax.fill_between(X_test.flatten(),
        mean_test.flatten() - range_shade * std_test.flatten(),
        mean_test.flatten() + range_shade * std_test.flatten(),
        color='b',
        alpha=0.3)

    if num_init is not None:
        if X_train.shape[0] > num_init:
            ax.plot(X_train[:num_init, :], Y_train[:num_init, :], 'x',
                c='saddlebrown', ms=14, markeredgewidth=6)
            ax.plot(X_train[num_init:X_train.shape[0]-1, :],
                Y_train[num_init:X_train.shape[0]-1, :], 'rx',
                ms=14, markeredgewidth=6)
            ax.plot(X_train[X_train.shape[0]-1, :],
                Y_train[X_train.shape[0]-1, :], c='orange', marker='+',
                ms=18, markeredgewidth=6)
        else:
            ax.plot(X_train, Y_train, 'x', c='saddlebrown', ms=14,
                markeredgewidth=6)
    else:
        ax.plot(X_train[:X_train.shape[0]-1, :],
            Y_train[:X_train.shape[0]-1, :], 'rx', ms=14, markeredgewidth=6)
        ax.plot(X_train[X_train.shape[0]-1, :],
            Y_train[X_train.shape[0]-1, :], c='orange', marker='+', ms=18,
            markeredgewidth=6)

    _set_ax_config(ax, str_x_axis, str_y_axis, xlim_min=np.min(X_test),
        xlim_max=np.max(X_test), draw_zero_axis=draw_zero_axis)

    plt.tight_layout()

    if path_save is not None and str_postfix is not None:
        _save_figure(path_save, str_postfix, str_prefix='bo_step_')
    _show_figure(pause_figure, time_pause)

@utils_common.validate_types
def plot_bo_step_with_acq(X_train: np.ndarray, Y_train: np.ndarray,
    X_test: np.ndarray, Y_test: np.ndarray, mean_test: np.ndarray,
    std_test: np.ndarray, acq_test: np.ndarray,
    path_save: constants.TYPING_UNION_STR_NONE=None,
    str_postfix: constants.TYPING_UNION_STR_NONE=None,
    str_x_axis: str='x',
    str_y_axis: str='y',
    str_acq_axis: str='acq.',
    num_init: constants.TYPING_UNION_INT_NONE=None,
    use_tex: bool=False,
    draw_zero_axis: bool=False,
    pause_figure: bool=True,
    time_pause: constants.TYPING_UNION_INT_FLOAT=constants.TIME_PAUSE,
    range_shade: float=constants.RANGE_SHADE,
) -> constants.TYPE_NONE: # pragma: no cover
    """
    It is for plotting Bayesian optimization results step by step.

    :param X_train: training inputs. Shape: (n, 1).
    :type X_train: numpy.ndarray
    :param Y_train: training outputs. Shape: (n, 1).
    :type Y_train: numpy.ndarray
    :param X_test: test inputs. Shape: (m, 1).
    :type X_test: numpy.ndarray
    :param Y_test: true test outputs. Shape: (m, 1).
    :type Y_test: numpy.ndarray
    :param mean_test: posterior predictive mean function values over `X_test`.
        Shape: (m, 1).
    :type mean_test: numpy.ndarray
    :param std_test: posterior predictive standard deviation function values
        over `X_test`. Shape: (m, 1).
    :type std_test: numpy.ndarray
    :param acq_test: acquisition funcion values over `X_test`. Shape: (m, 1).
    :type acq_test: numpy.ndarray
    :param path_save: None, or path for saving a figure.
    :type path_save: NoneType or str., optional
    :param str_postfix: None, or the name of postfix.
    :type str_postfix: NoneType or str., optional
    :param str_x_axis: the name of x axis.
    :type str_x_axis: str., optional
    :param str_y_axis: the name of y axis.
    :type str_y_axis: str., optional
    :param str_acq_axis: the name of acquisition function axis.
    :type str_acq_axis: str., optional
    :param num_init: None, or the number of initial examples.
    :type num_init: NoneType or int., optional
    :param use_tex: flag for using latex.
    :type use_tex: bool., optional
    :param draw_zero_axis: flag for drawing a zero axis.
    :type draw_zero_axis: bool., optional
    :param pause_figure: flag for pausing before closing a figure.
    :type pause_figure: bool., optional
    :param time_pause: pausing time.
    :type time_pause: int. or float, optional
    :param range_shade: shade range for standard deviation.
    :type range_shade: float, optional

    :returns: None.
    :rtype: NoneType

    :raises: AssertionError

    """

    assert isinstance(X_train, np.ndarray)
    assert isinstance(Y_train, np.ndarray)
    assert isinstance(X_test, np.ndarray)
    assert isinstance(Y_test, np.ndarray)
    assert isinstance(mean_test, np.ndarray)
    assert isinstance(std_test, np.ndarray)
    assert isinstance(path_save, (str, type(None)))
    assert isinstance(str_postfix, (str, type(None)))
    assert isinstance(str_x_axis, str)
    assert isinstance(str_y_axis, str)
    assert isinstance(str_acq_axis, str)
    assert isinstance(num_init, (int, type(None)))
    assert isinstance(use_tex, bool)
    assert isinstance(draw_zero_axis, bool)
    assert isinstance(pause_figure, bool)
    assert isinstance(time_pause, (int, float))
    assert isinstance(range_shade, float)
    assert len(X_train.shape) == 2
    assert len(X_test.shape) == 2
    assert len(Y_train.shape) == 2
    assert len(Y_test.shape) == 2
    assert len(mean_test.shape) == 2
    assert len(std_test.shape) == 2
    assert len(acq_test.shape) == 2
    assert X_train.shape[1] == X_test.shape[1] == 1
    assert Y_train.shape[1] == Y_test.shape[1] == 1
    assert X_train.shape[0] == Y_train.shape[0]
    assert mean_test.shape[1] == 1
    assert std_test.shape[1] == 1
    assert acq_test.shape[1] == 1
    assert X_test.shape[0] == Y_test.shape[0] == mean_test.shape[0] \
        == std_test.shape[0] == acq_test.shape[0]
    if num_init is not None:
        assert X_train.shape[0] >= num_init

    if plt is None or pylab is None:
        logger.info('matplotlib or pylab is not installed.')
        return
    _set_font_config(use_tex)

    _, (ax1, ax2) = plt.subplots(2, 1, gridspec_kw = {'height_ratios':[3, 1]})

    ax1.plot(X_test, Y_test, 'g', linewidth=4)
    ax1.plot(X_test, mean_test, 'b', linewidth=4)
    ax1.fill_between(X_test.flatten(),
        mean_test.flatten() - range_shade * std_test.flatten(),
        mean_test.flatten() + range_shade * std_test.flatten(),
        color='b',
        alpha=0.3)
    if num_init is not None:
        if X_train.shape[0] > num_init:
            ax1.plot(X_train[:num_init, :], Y_train[:num_init, :], 'x',
                c='saddlebrown', ms=14, markeredgewidth=6)
            ax1.plot(X_train[num_init:X_train.shape[0]-1, :],
                Y_train[num_init:X_train.shape[0]-1, :], 'rx',
                ms=14, markeredgewidth=6)
            ax1.plot(X_train[X_train.shape[0]-1, :],
                Y_train[X_train.shape[0]-1, :],
                c='orange', marker='+', ms=18, markeredgewidth=6)
        else:
            ax1.plot(X_train, Y_train, 'x', c='saddlebrown', ms=14, markeredgewidth=6)
    else:
        ax1.plot(X_train[:X_train.shape[0]-1, :],
            Y_train[:X_train.shape[0]-1, :], 'rx', ms=14, markeredgewidth=6)
        ax1.plot(X_train[X_train.shape[0]-1, :],
            Y_train[X_train.shape[0]-1, :], c='orange', marker='+',
            ms=18, markeredgewidth=6)

    _set_ax_config(ax1, None, str_y_axis, xlim_min=np.min(X_test),
        xlim_max=np.max(X_test), draw_zero_axis=draw_zero_axis)

    ax2.plot(X_test, acq_test, 'b', linewidth=4)
    _set_ax_config(ax2, str_x_axis, str_acq_axis, xlim_min=np.min(X_test),
        xlim_max=np.max(X_test), draw_zero_axis=draw_zero_axis)

    plt.tight_layout()

    if path_save is not None and str_postfix is not None:
        _save_figure(path_save, str_postfix, str_prefix='bo_step_acq_')
    _show_figure(pause_figure, time_pause)



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: November 5, 2020
#
"""It is utilities for Bayesian optimization."""

import numpy as np
try:
    from scipydirect import minimize as directminimize
except: # pragma: no cover
    directminimize = None
try:
    import cma
except: # pragma: no cover
    cma = None

from bayeso import acquisition
from bayeso import constants
from bayeso.utils import utils_covariance
from bayeso.utils import utils_common
from bayeso.utils import utils_logger

logger = utils_logger.get_logger('utils_bo')


@utils_common.validate_types
def get_best_acquisition_by_evaluation(initials: np.ndarray,
    fun_objective: constants.TYPING_CALLABLE
) -> np.ndarray:
    """
    It returns the best acquisition with respect to values of `fun_objective`.
    Here, the best acquisition is a minimizer of `fun_objective`.

    :param initials: inputs. Shape: (n, d).
    :type initials: numpy.ndarray
    :param fun_objective: an objective function.
    :type fun_objective: callable

    :returns: the best example of `initials`. Shape: (1, d).
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(initials, np.ndarray)
    assert callable(fun_objective)
    assert len(initials.shape) == 2

    acq_val_best = np.inf
    initial_best = None
    for initial in initials:
        acq_val = fun_objective(initial)
        if acq_val < acq_val_best:
            initial_best = initial
            acq_val_best = acq_val
    initial_best = initial_best[np.newaxis, ...]
    return initial_best

@utils_common.validate_types
def get_best_acquisition_by_history(X: np.ndarray, Y: np.ndarray
) -> constants.TYPING_TUPLE_ARRAY_FLOAT:
    """
    It returns the best acquisition that has shown minimum result, and its minimum result.

    :param X: historical queries. Shape: (n, d).
    :type X: numpy.ndarray
    :param Y: the observations of `X`. Shape: (n, 1).
    :type Y: numpy.ndarray

    :returns: a tuple of the best query and its result.
    :rtype: (numpy.ndarray, float)

    :raises: AssertionError

    """

    assert isinstance(X, np.ndarray)
    assert isinstance(Y, np.ndarray)
    assert len(X.shape) == 2
    assert len(Y.shape) == 2
    assert X.shape[0] == Y.shape[0]
    assert Y.shape[1] == 1

    ind_best = np.argmin(Y)
    bx_best = X[ind_best]
    y_best = Y[ind_best, 0]

    return bx_best, y_best

@utils_common.validate_types
def get_next_best_acquisition(points: np.ndarray, acquisitions: np.ndarray,
    points_evaluated: np.ndarray
) -> np.ndarray:
    """
    It returns the next best acquired sample.

    :param points: inputs for acquisition function. Shape: (n, d).
    :type points: numpy.ndarray
    :param acquisitions: acquisition function values over `points`. Shape: (n, ).
    :type acquisitions: numpy.ndarray
    :param points_evaluated: the samples evaluated so far. Shape: (m, d).
    :type points_evaluated: numpy.ndarray

    :returns: next best acquired point. Shape: (d, ).
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(points, np.ndarray)
    assert isinstance(acquisitions, np.ndarray)
    assert isinstance(points_evaluated, np.ndarray)
    assert len(points.shape) == 2
    assert len(acquisitions.shape) == 1
    assert len(points_evaluated.shape) == 2
    assert points.shape[0] == acquisitions.shape[0]
    assert points.shape[1] == points_evaluated.shape[1]

    for cur_point in points_evaluated:
        ind_same, = np.where(np.linalg.norm(points - cur_point, axis=1) < 1e-2)
        points = np.delete(points, ind_same, axis=0)
        acquisitions = np.delete(acquisitions, ind_same)
    cur_best = np.inf
    next_point = None

    if points.shape[0] > 0:
        for arr_point, cur_acq in zip(points, acquisitions):
            if cur_acq < cur_best:
                cur_best = cur_acq
                next_point = arr_point
    else:
        next_point = points_evaluated[-1]
    return next_point

@utils_common.validate_types
def check_optimizer_method_bo(str_optimizer_method_bo: str, dim: int, debug: bool) -> str:
    """
    It checks the availability of optimization methods.
    It helps to run Bayesian optimization, even though additional
    optimization methods are not installed or there exist the conditions
    some of optimization methods cannot be run.

    :param str_optimizer_method_bo: the name of optimization method for
        Bayesian optimization.
    :type str_optimizer_method_bo: str.
    :param dim: dimensionality of the problem we solve.
    :type dim: int.
    :param debug: flag for printing log messages.
    :type debug: bool.

    :returns: available `str_optimizer_method_bo`.
    :rtype: str.

    :raises: AssertionError

    """

    assert isinstance(str_optimizer_method_bo, str)
    assert isinstance(dim, int)
    assert isinstance(debug, bool)
    assert str_optimizer_method_bo in constants.ALLOWED_OPTIMIZER_METHOD_BO \
        + constants.ALLOWED_OPTIMIZER_METHOD_BO_TREES

    if str_optimizer_method_bo == 'DIRECT' and directminimize is None: # pragma: no cover
        logger.warning('DIRECT is selected, but it is not installed.')
        str_optimizer_method_bo = 'L-BFGS-B'
    elif str_optimizer_method_bo == 'CMA-ES' and cma is None: # pragma: no cover
        logger.warning('CMA-ES is selected, but it is not installed.')
        str_optimizer_method_bo = 'L-BFGS-B'
    # TODO: It should be checked.
    elif str_optimizer_method_bo == 'CMA-ES' and dim == 1: # pragma: no cover
        logger.warning('CMA-ES is selected, but a dimension of bounds is 1.')
        str_optimizer_method_bo = 'L-BFGS-B'
    return str_optimizer_method_bo

@utils_common.validate_types
def choose_fun_acquisition(
    str_acq: str, noise: constants.TYPING_UNION_FLOAT_NONE=None
) -> constants.TYPING_CALLABLE:
    """
    It chooses and returns an acquisition function.

    :param str_acq: the name of acquisition function.
    :type str_acq: str.
    :param hyps: dictionary of hyperparameters for acquisition function.
    :type hyps: dict.

    :returns: acquisition function.
    :rtype: callable

    :raises: AssertionError

    """

    assert isinstance(str_acq, str)
    assert isinstance(noise, (float, constants.TYPE_NONE))
    assert str_acq in constants.ALLOWED_BO_ACQ

    if str_acq == 'pi':
        fun_acquisition = acquisition.pi
    elif str_acq == 'ei':
        fun_acquisition = acquisition.ei
    elif str_acq == 'ucb':
        fun_acquisition = acquisition.ucb
    elif str_acq == 'aei':
        assert noise is not None

        fun_acquisition = lambda pred_mean, pred_std, Y_train: acquisition.aei(
            pred_mean, pred_std, Y_train, noise)
    elif str_acq == 'pure_exploit':
        fun_acquisition = lambda pred_mean, pred_std, Y_train: acquisition.pure_exploit(
            pred_mean)
    elif str_acq == 'pure_explore':
        fun_acquisition = lambda pred_mean, pred_std, Y_train: acquisition.pure_explore(pred_std)
    else:
        raise NotImplementedError('_choose_fun_acquisition: allowed str_acq,\
            but it is not implemented.')

    return fun_acquisition

@utils_common.validate_types
def check_hyps_convergence(list_hyps: constants.TYPING_LIST[dict], hyps: dict,
    str_cov: str, fix_noise: bool,
    ratio_threshold: float=0.05
) -> bool:
    """
    It checks convergence of hyperparameters for Gaussian process regression.

    :param list_hyps: list of historical hyperparameters for Gaussian process regression.
    :type list_hyps: list
    :param hyps: dictionary of hyperparameters for acquisition function.
    :type hyps: dict.
    :param str_cov: the name of covariance function.
    :type str_cov: str.
    :param fix_noise: flag for fixing a noise.
    :type fix_noise: bool.
    :param ratio_threshold: ratio of threshold for checking convergence.
    :type ratio_threshold: float, optional

    :returns: flag for checking convergence. If converged, it is True.
    :rtype: bool.

    :raises: AssertionError

    """

    assert isinstance(list_hyps, list)
    assert isinstance(hyps, dict)
    assert isinstance(str_cov, str)
    assert isinstance(fix_noise, bool)
    assert isinstance(ratio_threshold, float)

    converged = False
    if len(list_hyps) > 0:
        hyps_converted = utils_covariance.convert_hyps(str_cov, hyps, fix_noise=fix_noise)
        target_hyps_converted = utils_covariance.convert_hyps(str_cov, list_hyps[-1],
            fix_noise=fix_noise)

        threshold = np.linalg.norm(target_hyps_converted) * ratio_threshold
        if np.linalg.norm(hyps_converted - target_hyps_converted, ord=2) < threshold:
            converged = True
    return converged



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: March 22, 2021
#
"""These files are for implementing Gaussian process regression.
It is implemented, based on the following article:

(i) Rasmussen, C. E., & Williams, C. K. (2006). Gaussian Process
Regression for Machine Learning. MIT Press."""



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: October 8, 2021
#
"""It defines a class of Bayesian optimization
with tree-based surrogate models."""

import numpy as np
import time

from bayeso.bo import base_bo
from bayeso.trees import trees_common
from bayeso import constants
from bayeso.utils import utils_bo
from bayeso.utils import utils_logger


class BOwTrees(base_bo.BaseBO):
    """
    It is a Bayesian optimization class with tree-based surrogate models.

    :param range_X: a search space. Shape: (d, 2).
    :type range_X: numpy.ndarray
    :param str_surrogate: the name of surrogate model.
    :type str_surrogate: str., optional
    :param str_acq: the name of acquisition function.
    :type str_acq: str., optional
    :param normalize_Y: flag for normalizing outputs.
    :type normalize_Y: bool., optional
    :param str_optimizer_method_bo: the name of optimization method for
        Bayesian optimization.
    :type str_optimizer_method_bo: str., optional
    :param debug: flag for printing log messages.
    :type debug: bool., optional

    """

    def __init__(self, range_X: np.ndarray,
        str_surrogate: str=constants.STR_SURROGATE_TREES,
        str_acq: str=constants.STR_BO_ACQ,
        normalize_Y: bool=constants.NORMALIZE_RESPONSE,
        str_optimizer_method_bo: str=constants.STR_OPTIMIZER_METHOD_AO_TREES,
        debug: bool=False
    ):
        """
        Constructor method

        """

        assert isinstance(range_X, np.ndarray)
        assert isinstance(str_surrogate, str)
        assert isinstance(str_acq, str)
        assert isinstance(normalize_Y, bool)
        assert isinstance(str_optimizer_method_bo, str)
        assert isinstance(debug, bool)
        assert len(range_X.shape) == 2
        assert range_X.shape[1] == 2
        assert (range_X[:, 0] <= range_X[:, 1]).all()
        assert str_surrogate in constants.ALLOWED_SURROGATE_TREES
        assert str_acq in constants.ALLOWED_BO_ACQ
        assert str_optimizer_method_bo in constants.ALLOWED_OPTIMIZER_METHOD_BO_TREES

        super().__init__(range_X, str_surrogate, str_acq, str_optimizer_method_bo, normalize_Y, debug)

    def get_trees(self, X_train, Y_train,
        num_trees=100,
        depth_max=5,
        size_min_leaf=1,
    ):
        """
        It returns a list of trees.

        :param X_train: inputs. Shape: (n, d).
        :type X_train: numpy.ndarray
        :param Y_train: outputs. Shape: (n, 1).
        :type Y_train: numpy.ndarray
        :param num_trees: the number of trees.
        :type num_trees: int., optional
        :param depth_max: maximum depth.
        :type depth_max: int., optional
        :param size_min_leaf: minimum size of leaves.
        :type size_min_leaf: int., optional

        :returns: list of trees.
        :rtype: list

        :raises: AssertionError

        """

        assert isinstance(X_train, np.ndarray)
        assert isinstance(Y_train, np.ndarray)
        assert isinstance(num_trees, int)
        assert isinstance(depth_max, int)
        assert isinstance(size_min_leaf, int)
        assert len(X_train.shape) == 2
        assert len(Y_train.shape) == 2
        assert X_train.shape[0] == Y_train.shape[0]
        assert Y_train.shape[1] == 1

        num_features = int(np.sqrt(self.num_dim))

        if self.str_surrogate == 'rf':
            from bayeso.trees import trees_random_forest
            trees = trees_random_forest.get_random_forest(
                X_train, Y_train, num_trees, depth_max, size_min_leaf, num_features
            )
        else:
            raise NotImplementedError('allowed str_surrogate, but it is not implemented.')

        return trees

    def compute_posteriors(self,
        X: np.ndarray, trees: constants.TYPING_LIST
    ) -> np.ndarray:
        """
        It returns posterior mean and standard deviation functions over `X`.

        :param X: inputs to test. Shape: (l, d).
        :type X: numpy.ndarray
        :param trees: list of trees.
        :type trees: list

        :returns: posterior mean and standard deviation functions
            over `X`. Shape: ((l, ), (l, )).
        :rtype: (numpy.ndarray, numpy.ndarray)

        :raises: AssertionError

        """

        assert isinstance(X, np.ndarray)
        assert isinstance(trees, list)
        assert len(X.shape) == 2
        assert X.shape[1] == self.num_dim

        pred_mean, pred_std = trees_common.predict_by_trees(X, trees)

        pred_mean = np.squeeze(pred_mean, axis=1)
        pred_std = np.squeeze(pred_std, axis=1)

        return pred_mean, pred_std

    def compute_acquisitions(self, X: np.ndarray,
        X_train: np.ndarray, Y_train: np.ndarray,
        trees: constants.TYPING_LIST
    ) -> np.ndarray:
        """
        It computes acquisition function values over 'X',
        where `X_train` and `Y_train` are given.

        :param X: inputs. Shape: (l, d).
        :type X: numpy.ndarray
        :param X_train: inputs. Shape: (n, d).
        :type X_train: numpy.ndarray
        :param Y_train: outputs. Shape: (n, 1).
        :type Y_train: numpy.ndarray
        :param trees: list of trees.
        :type trees: list

        :returns: acquisition function values over `X`. Shape: (l, ).
        :rtype: numpy.ndarray

        :raises: AssertionError

        """

        assert isinstance(X, np.ndarray)
        assert isinstance(X_train, np.ndarray)
        assert isinstance(Y_train, np.ndarray)
        assert len(X.shape) == 1 or len(X.shape) == 2 or len(X.shape) == 3
        assert len(X_train.shape) == 2 or len(X_train.shape) == 3
        assert len(Y_train.shape) == 2
        assert Y_train.shape[1] == 1
        assert X_train.shape[0] == Y_train.shape[0]

        if len(X.shape) == 1:
            X = np.atleast_2d(X)

        assert X.shape[1] == X_train.shape[1] == self.num_dim

        fun_acquisition = utils_bo.choose_fun_acquisition(self.str_acq, constants.GP_NOISE)

        pred_mean, pred_std = self.compute_posteriors(X, trees)

        acquisitions = fun_acquisition(
            pred_mean=pred_mean, pred_std=pred_std, Y_train=Y_train
        )
        acquisitions *= constants.MULTIPLIER_ACQ

        return acquisitions

    def optimize(self, X_train: np.ndarray, Y_train: np.ndarray,
        str_sampling_method: str=constants.STR_SAMPLING_METHOD_AO_TREES,
        num_samples: int=constants.NUM_SAMPLES_AO_TREES,
    ) -> constants.TYPING_TUPLE_ARRAY_DICT:
        """
        It computes acquired example, candidates of acquired examples,
        acquisition function values for the candidates, covariance matrix,
        inverse matrix of the covariance matrix, hyperparameters optimized,
        and execution times.

        :param X_train: inputs. Shape: (n, d).
        :type X_train: numpy.ndarray
        :param Y_train: outputs. Shape: (n, 1).
        :type Y_train: numpy.ndarray
        :param str_sampling_method: the name of sampling method for
            acquisition function optimization.
        :type str_sampling_method: str., optional
        :param num_samples: the number of samples.
        :type num_samples: int., optional

        :returns: acquired example and dictionary of information. Shape: ((d, ), dict.).
        :rtype: (numpy.ndarray, dict.)

        :raises: AssertionError

        """

        assert isinstance(X_train, np.ndarray)
        assert isinstance(Y_train, np.ndarray)
        assert isinstance(str_sampling_method, str)
        assert isinstance(num_samples, int)
        assert len(X_train.shape) == 2
        assert len(Y_train.shape) == 2
        assert Y_train.shape[1] == 1
        assert X_train.shape[0] == Y_train.shape[0]
        assert X_train.shape[1] == self.num_dim
        assert num_samples > 0
        assert str_sampling_method in constants.ALLOWED_SAMPLING_METHOD

        time_start = time.time()

        if self.normalize_Y and np.max(Y_train) != np.min(Y_train):
            Y_train = (Y_train - np.min(Y_train)) / (np.max(Y_train) - np.min(Y_train)) \
                * constants.MULTIPLIER_RESPONSE

        time_start_surrogate = time.time()
        trees = self.get_trees(X_train, Y_train)
        time_end_surrogate = time.time()

        time_start_acq = time.time()
        next_points = self.get_samples(
            str_sampling_method,
            fun_objective=None,
            num_samples=num_samples
        )

        fun_negative_acquisition = lambda X_test: -1.0 * self.compute_acquisitions(
            X_test, X_train, Y_train, trees
        )
        acquisitions = fun_negative_acquisition(next_points)
        ind_next_point = np.argmin(acquisitions)
        next_point = next_points[ind_next_point]

        time_end_acq = time.time()

        time_end = time.time()

        dict_info = {
            'next_points': next_points,
            'acquisitions': acquisitions,
            'trees': trees,
            'time_surrogate': time_end_surrogate - time_start_surrogate,
            'time_acq': time_end_acq - time_start_acq,
            'time_overall': time_end - time_start,
        }

        if self.debug:
            self.logger.debug('overall time consumed to acquire: %.4f sec.', time_end - time_start)

        return next_point, dict_info



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: March 22, 2021
#
"""It defines the functions related to likelihood for
Gaussian process regression."""

import numpy as np
import scipy.linalg

from bayeso import covariance
from bayeso import constants
from bayeso.utils import utils_gp
from bayeso.utils import utils_covariance
from bayeso.utils import utils_common
from bayeso.utils import utils_logger

logger = utils_logger.get_logger('gp_likelihood')


@utils_common.validate_types
def neg_log_ml(X_train: np.ndarray, Y_train: np.ndarray, hyps: np.ndarray,
    str_cov: str, prior_mu_train: np.ndarray,
    use_ard: bool=constants.USE_ARD,
    fix_noise: bool=constants.FIX_GP_NOISE,
    use_cholesky: bool=True,
    use_gradient: bool=True,
    debug: bool=False
) -> constants.TYPING_UNION_FLOAT_FA:
    """
    This function computes a negative log marginal likelihood.

    :param X_train: inputs. Shape: (n, d) or (n, m, d).
    :type X_train: numpy.ndarray
    :param Y_train: outputs. Shape: (n, 1).
    :type Y_train: numpy.ndarray
    :param hyps: hyperparameters for Gaussian process. Shape: (h, ).
    :type hyps: numpy.ndarray
    :param str_cov: the name of covariance function.
    :type str_cov: str.
    :param prior_mu_train: the prior values computed by get_prior_mu(). Shape: (n, 1).
    :type prior_mu_train: numpy.ndarray
    :param use_ard: flag for automatic relevance determination.
    :type use_ard: bool., optional
    :param fix_noise: flag for fixing a noise.
    :type fix_noise: bool., optional
    :param use_cholesky: flag for using a cholesky decomposition.
    :type use_cholesky: bool., optional
    :param use_gradient: flag for computing and returning gradients of
        negative log marginal likelihood.
    :type use_gradient: bool., optional
    :param debug: flag for printing log messages.
    :type debug: bool., optional

    :returns: negative log marginal likelihood, or (negative log marginal
        likelihood, gradients of the likelihood).
    :rtype: float, or tuple of (float, np.ndarray)

    :raises: AssertionError

    """

    # TODO: add use_ard.
    utils_gp.validate_common_args(X_train, Y_train, str_cov, None, debug)
    assert isinstance(hyps, np.ndarray)
    assert isinstance(prior_mu_train, np.ndarray)
    assert isinstance(use_ard, bool)
    assert isinstance(fix_noise, bool)
    assert isinstance(use_cholesky, bool)
    assert isinstance(use_gradient, bool)
    assert len(prior_mu_train.shape) == 2
    assert X_train.shape[0] == Y_train.shape[0] == prior_mu_train.shape[0]
    utils_covariance.check_str_cov('neg_log_ml', str_cov, X_train.shape)

    hyps = utils_covariance.restore_hyps(str_cov, hyps, use_ard=use_ard, fix_noise=fix_noise)
    new_Y_train = Y_train - prior_mu_train
    if use_cholesky:
        cov_X_X, lower, grad_cov_X_X = covariance.get_kernel_cholesky(X_train,
            hyps, str_cov, fix_noise=fix_noise, use_gradient=use_gradient,
            debug=debug)

        alpha = scipy.linalg.cho_solve((lower, True), new_Y_train)

        first_term = -0.5 * np.dot(new_Y_train.T, alpha)
        second_term = -1.0 * np.sum(np.log(np.diagonal(lower) + constants.JITTER_LOG))

        if use_gradient:
            assert grad_cov_X_X is not None

            first_term_grad = np.einsum("ik,jk->ijk", alpha, alpha)
            first_term_grad -= np.expand_dims(scipy.linalg.cho_solve((lower, True),
                np.eye(cov_X_X.shape[0])), axis=2)
            grad_log_ml_ = 0.5 * np.einsum("ijl,ijk->kl", first_term_grad, grad_cov_X_X)
            grad_log_ml_ = np.sum(grad_log_ml_, axis=1)
    else:
        # TODO: use_gradient is fixed.
        use_gradient = False
        cov_X_X, inv_cov_X_X, grad_cov_X_X = covariance.get_kernel_inverse(X_train,
            hyps, str_cov, fix_noise=fix_noise, use_gradient=use_gradient,
            debug=debug)

        first_term = -0.5 * np.dot(np.dot(new_Y_train.T, inv_cov_X_X), new_Y_train)
        sign_second_term, second_term = np.linalg.slogdet(cov_X_X)

        # TODO: It should be checked.
        if sign_second_term <= 0: # pragma: no cover
            second_term = 0.0

        second_term = -0.5 * second_term

    third_term = -float(X_train.shape[0]) / 2.0 * np.log(2.0 * np.pi)
    log_ml_ = np.squeeze(first_term + second_term + third_term)
    log_ml_ /= X_train.shape[0]

    if use_gradient:
        return -1.0 * log_ml_, -1.0 * grad_log_ml_ / X_train.shape[0]

    return -1.0 * log_ml_

@utils_common.validate_types
def neg_log_pseudo_l_loocv(X_train: np.ndarray, Y_train: np.ndarray, hyps: np.ndarray,
    str_cov: str, prior_mu_train: np.ndarray,
    fix_noise: bool=constants.FIX_GP_NOISE,
    debug: bool=False
) -> float:
    """
    It computes a negative log pseudo-likelihood using leave-one-out cross-validation.

    :param X_train: inputs. Shape: (n, d) or (n, m, d).
    :type X_train: numpy.ndarray
    :param Y_train: outputs. Shape: (n, 1).
    :type Y_train: numpy.ndarray
    :param hyps: hyperparameters for Gaussian process. Shape: (h, ).
    :type hyps: numpy.ndarray
    :param str_cov: the name of covariance function.
    :type str_cov: str.
    :param prior_mu_train: the prior values computed by get_prior_mu(). Shape: (n, 1).
    :type prior_mu_train: numpy.ndarray
    :param fix_noise: flag for fixing a noise.
    :type fix_noise: bool., optional
    :param debug: flag for printing log messages.
    :type debug: bool., optional

    :returns: negative log pseudo-likelihood.
    :rtype: float

    :raises: AssertionError

    """

    # TODO: add use_ard.
    utils_gp.validate_common_args(X_train, Y_train, str_cov, None, debug)
    assert isinstance(hyps, np.ndarray)
    assert isinstance(prior_mu_train, np.ndarray)
    assert isinstance(fix_noise, bool)
    assert len(prior_mu_train.shape) == 2
    assert X_train.shape[0] == Y_train.shape[0] == prior_mu_train.shape[0]
    utils_covariance.check_str_cov('neg_log_pseudo_l_loocv', str_cov, X_train.shape)

    num_data = X_train.shape[0]
    hyps = utils_covariance.restore_hyps(str_cov, hyps, fix_noise=fix_noise)

    _, inv_cov_X_X, _ = covariance.get_kernel_inverse(X_train, hyps,
        str_cov, fix_noise=fix_noise, debug=debug)

    log_pseudo_l_ = 0.0
    for ind_data in range(0, num_data):
        # TODO: check this.
#        cur_X_train = np.vstack((X_train[:ind_data], X_train[ind_data+1:]))
#        cur_Y_train = np.vstack((Y_train[:ind_data], Y_train[ind_data+1:]))

#        cur_X_test = np.expand_dims(X_train[ind_data], axis=0)
        cur_Y_test = Y_train[ind_data]

        cur_mu = np.squeeze(cur_Y_test) \
            - np.dot(inv_cov_X_X, Y_train)[ind_data] / inv_cov_X_X[ind_data, ind_data]
        cur_sigma = np.sqrt(1.0 / (inv_cov_X_X[ind_data, ind_data] + constants.JITTER_COV))

        first_term = -0.5 * np.log(cur_sigma**2)
        second_term = -0.5 * (np.squeeze(cur_Y_test - cur_mu))**2 / (cur_sigma**2)
        third_term = -0.5 * np.log(2.0 * np.pi)
        cur_log_pseudo_l_ = first_term + second_term + third_term
        log_pseudo_l_ += cur_log_pseudo_l_

    log_pseudo_l_ /= num_data
    log_pseudo_l_ *= -1.0

    return log_pseudo_l_



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: March 22, 2021
#
"""It defines the functions related to kernels for
Gaussian process regression."""

import time
import numpy as np
import scipy.optimize

from bayeso import covariance
from bayeso import constants
from bayeso.gp import gp_likelihood
from bayeso.utils import utils_gp
from bayeso.utils import utils_covariance
from bayeso.utils import utils_common
from bayeso.utils import utils_logger

logger = utils_logger.get_logger('gp_kernel')


@utils_common.validate_types
def get_optimized_kernel(X_train: np.ndarray, Y_train: np.ndarray,
    prior_mu: constants.TYPING_UNION_CALLABLE_NONE, str_cov: str,
    str_optimizer_method: str=constants.STR_OPTIMIZER_METHOD_GP,
    str_modelselection_method: str=constants.STR_MODELSELECTION_METHOD,
    use_ard: bool=constants.USE_ARD,
    fix_noise: bool=constants.FIX_GP_NOISE,
    debug: bool=False
) -> constants.TYPING_TUPLE_TWO_ARRAYS_DICT:
    """
    This function computes the kernel matrix optimized by optimization
    method specified, its inverse matrix, and the optimized hyperparameters.

    :param X_train: inputs. Shape: (n, d) or (n, m, d).
    :type X_train: numpy.ndarray
    :param Y_train: outputs. Shape: (n, 1).
    :type Y_train: numpy.ndarray
    :param prior_mu: prior mean function or None.
    :type prior_mu: callable or NoneType
    :param str_cov: the name of covariance function.
    :type str_cov: str.
    :param str_optimizer_method: the name of optimization method.
    :type str_optimizer_method: str., optional
    :param str_modelselection_method: the name of model selection method.
    :type str_modelselection_method: str., optional
    :param use_ard: flag for using automatic relevance determination.
    :type use_ard: bool., optional
    :param fix_noise: flag for fixing a noise.
    :type fix_noise: bool., optional
    :param debug: flag for printing log messages.
    :type debug: bool., optional

    :returns: a tuple of kernel matrix over `X_train`, kernel matrix
        inverse, and dictionary of hyperparameters.
    :rtype: tuple of (numpy.ndarray, numpy.ndarray, dict.)

    :raises: AssertionError, ValueError

    """

    # TODO: check to input same fix_noise to convert_hyps and restore_hyps
    utils_gp.validate_common_args(X_train, Y_train, str_cov, prior_mu, debug)
    assert isinstance(str_optimizer_method, str)
    assert isinstance(str_modelselection_method, str)
    assert isinstance(use_ard, bool)
    assert isinstance(fix_noise, bool)
    utils_covariance.check_str_cov('get_optimized_kernel', str_cov, X_train.shape)
    assert str_optimizer_method in constants.ALLOWED_OPTIMIZER_METHOD_GP
    assert str_modelselection_method in constants.ALLOWED_MODELSELECTION_METHOD
    use_gradient = bool(str_optimizer_method != 'Nelder-Mead')
    # TODO: Now, use_gradient is fixed as False.
#    use_gradient = False

    time_start = time.time()

    if debug:
        logger.debug('str_optimizer_method: %s', str_optimizer_method)
        logger.debug('str_modelselection_method: %s', str_modelselection_method)
        logger.debug('use_gradient: %s', use_gradient)

    prior_mu_train = utils_gp.get_prior_mu(prior_mu, X_train)
    if str_cov in constants.ALLOWED_COV_BASE:
        num_dim = X_train.shape[1]
    elif str_cov in constants.ALLOWED_COV_SET:
        num_dim = X_train.shape[2]
        use_gradient = False

    if str_modelselection_method == 'ml':
        neg_log_ml_ = lambda hyps: gp_likelihood.neg_log_ml(X_train, Y_train,
            hyps, str_cov, prior_mu_train,
            use_ard=use_ard, fix_noise=fix_noise,
            use_gradient=use_gradient, debug=debug)
    elif str_modelselection_method == 'loocv':
        # TODO: add use_ard.
        neg_log_ml_ = lambda hyps: gp_likelihood.neg_log_pseudo_l_loocv(X_train, Y_train,
            hyps, str_cov, prior_mu_train, fix_noise=fix_noise, debug=debug)
        use_gradient = False
    else: # pragma: no cover
        raise ValueError('get_optimized_kernel: missing conditions for str_modelselection_method.')

    hyps_converted = utils_covariance.convert_hyps(
        str_cov,
        utils_covariance.get_hyps(str_cov, num_dim, use_ard=use_ard),
        fix_noise=fix_noise
    )

    if str_optimizer_method in ['BFGS', 'SLSQP']:
        result_optimized = scipy.optimize.minimize(neg_log_ml_, hyps_converted,
            method=str_optimizer_method, jac=use_gradient, options={'disp': False})

        if debug:
            logger.debug('negative log marginal likelihood: %.6f', result_optimized.fun)
            logger.debug('scipy message: %s', result_optimized.message)

        result_optimized = result_optimized.x
    elif str_optimizer_method in ['L-BFGS-B', 'SLSQP-Bounded']:
        if str_optimizer_method == 'SLSQP-Bounded':
            str_optimizer_method = 'SLSQP'

        bounds = utils_covariance.get_range_hyps(str_cov, num_dim,
            use_ard=use_ard,
            fix_noise=fix_noise)
        result_optimized = scipy.optimize.minimize(neg_log_ml_, hyps_converted,
            method=str_optimizer_method, bounds=bounds, jac=use_gradient,
            options={'disp': False})

        if debug:
            logger.debug('negative log marginal likelihood: %.6f', result_optimized.fun)
            logger.debug('scipy message: %s', result_optimized.message)
        result_optimized = result_optimized.x
    elif str_optimizer_method in ['Nelder-Mead']:
        result_optimized = scipy.optimize.minimize(neg_log_ml_, hyps_converted,
            method=str_optimizer_method, options={'disp': False})

        if debug:
            logger.debug('negative log marginal likelihood: %.6f', result_optimized.fun)
            logger.debug('scipy message: %s', result_optimized.message)
        result_optimized = result_optimized.x
    else: # pragma: no cover
        raise ValueError('get_optimized_kernel: missing conditions for str_optimizer_method')

    hyps = utils_covariance.restore_hyps(str_cov, result_optimized,
        use_ard=use_ard,
        fix_noise=fix_noise)

    hyps = utils_covariance.validate_hyps_dict(hyps, str_cov, num_dim)
    cov_X_X, inv_cov_X_X, _ = covariance.get_kernel_inverse(X_train,
        hyps, str_cov, fix_noise=fix_noise, debug=debug)
    time_end = time.time()

    if debug:
        logger.debug('hyps optimized: %s', utils_logger.get_str_hyps(hyps))
        logger.debug('time consumed to construct gpr: %.4f sec.', time_end - time_start)
    return cov_X_X, inv_cov_X_X, hyps



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: September 24, 2020
#
"""It defines acquisition functions, each of which
is employed to determine where next to evaluate."""

import numpy as np
import scipy.stats

from bayeso import constants
from bayeso.utils import utils_common


@utils_common.validate_types
def pi(pred_mean: np.ndarray, pred_std: np.ndarray, Y_train: np.ndarray,
    jitter: float=constants.JITTER_ACQ
) -> np.ndarray:
    """
    It is a probability of improvement criterion.

    :param pred_mean: posterior predictive mean function over `X_test`.
        Shape: (l, ).
    :type pred_mean: numpy.ndarray
    :param pred_std: posterior predictive standard deviation function over
        `X_test`. Shape: (l, ).
    :type pred_std: numpy.ndarray
    :param Y_train: outputs of `X_train`. Shape: (n, 1).
    :type Y_train: numpy.ndarray
    :param jitter: jitter for `pred_std`.
    :type jitter: float, optional

    :returns: acquisition function values. Shape: (l, ).
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(pred_mean, np.ndarray)
    assert isinstance(pred_std, np.ndarray)
    assert isinstance(Y_train, np.ndarray)
    assert isinstance(jitter, float)
    assert len(pred_mean.shape) == 1
    assert len(pred_std.shape) == 1
    assert len(Y_train.shape) == 2
    assert pred_mean.shape[0] == pred_std.shape[0]

    with np.errstate(divide='ignore'):
        val_z = (np.min(Y_train) - pred_mean) / (pred_std + jitter)
    return scipy.stats.norm.cdf(val_z)

@utils_common.validate_types
def ei(pred_mean: np.ndarray, pred_std: np.ndarray, Y_train: np.ndarray,
    jitter: float=constants.JITTER_ACQ
) -> np.ndarray:
    """
    It is an expected improvement criterion.

    :param pred_mean: posterior predictive mean function over `X_test`. Shape: (l, ).
    :type pred_mean: numpy.ndarray
    :param pred_std: posterior predictive standard deviation function over `X_test`. Shape: (l, ).
    :type pred_std: numpy.ndarray
    :param Y_train: outputs of `X_train`. Shape: (n, 1).
    :type Y_train: numpy.ndarray
    :param jitter: jitter for `pred_std`.
    :type jitter: float, optional

    :returns: acquisition function values. Shape: (l, ).
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(pred_mean, np.ndarray)
    assert isinstance(pred_std, np.ndarray)
    assert isinstance(Y_train, np.ndarray)
    assert isinstance(jitter, float)
    assert len(pred_mean.shape) == 1
    assert len(pred_std.shape) == 1
    assert len(Y_train.shape) == 2
    assert pred_mean.shape[0] == pred_std.shape[0]

    with np.errstate(divide='ignore'):
        val_z = (np.min(Y_train) - pred_mean) / (pred_std + jitter)
    return (np.min(Y_train) - pred_mean) * scipy.stats.norm.cdf(val_z) \
        + pred_std * scipy.stats.norm.pdf(val_z)

@utils_common.validate_types
def ucb(pred_mean: np.ndarray, pred_std: np.ndarray,
    Y_train: constants.TYPING_UNION_ARRAY_NONE=None,
    kappa: float=2.0,
    increase_kappa: bool=True
) -> np.ndarray:
    """
    It is a Gaussian process upper confidence bound criterion.

    :param pred_mean: posterior predictive mean function over `X_test`.
        Shape: (l, ).
    :type pred_mean: numpy.ndarray
    :param pred_std: posterior predictive standard deviation function over
        `X_test`. Shape: (l, ).
    :type pred_std: numpy.ndarray
    :param Y_train: outputs of `X_train`. Shape: (n, 1).
    :type Y_train: numpy.ndarray, optional
    :param kappa: trade-off hyperparameter between exploration and
        exploitation.
    :type kappa: float, optional
    :param increase_kappa: flag for increasing a kappa value as `Y_train`
        grows. If `Y_train` is None, it is ignored, which means `kappa` is
        fixed.
    :type increase_kappa: bool., optional

    :returns: acquisition function values. Shape: (l, ).
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(pred_mean, np.ndarray)
    assert isinstance(pred_std, np.ndarray)
    assert isinstance(Y_train, (np.ndarray, type(None)))
    assert isinstance(kappa, float)
    assert isinstance(increase_kappa, bool)
    assert len(pred_mean.shape) == 1
    assert len(pred_std.shape) == 1
    if Y_train is not None:
        assert len(Y_train.shape) == 2
    assert pred_mean.shape[0] == pred_std.shape[0]

    if increase_kappa and Y_train is not None:
        kappa_ = kappa * np.log(Y_train.shape[0])
    else:
        kappa_ = kappa
    return -pred_mean + kappa_ * pred_std

@utils_common.validate_types
def aei(pred_mean: np.ndarray, pred_std: np.ndarray, Y_train: np.ndarray, noise: float,
    jitter: float=constants.JITTER_ACQ
) -> np.ndarray:
    """
    It is an augmented expected improvement criterion.

    :param pred_mean: posterior predictive mean function over `X_test`. Shape: (l, ).
    :type pred_mean: numpy.ndarray
    :param pred_std: posterior predictive standard deviation function over `X_test`. Shape: (l, ).
    :type pred_std: numpy.ndarray
    :param Y_train: outputs of `X_train`. Shape: (n, 1).
    :type Y_train: numpy.ndarray
    :param noise: noise for augmenting exploration.
    :type noise: float
    :param jitter: jitter for `pred_std`.
    :type jitter: float, optional

    :returns: acquisition function values. Shape: (l, ).
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(pred_mean, np.ndarray)
    assert isinstance(pred_std, np.ndarray)
    assert isinstance(Y_train, np.ndarray)
    assert isinstance(noise, float)
    assert isinstance(jitter, float)
    assert len(pred_mean.shape) == 1
    assert len(pred_std.shape) == 1
    assert len(Y_train.shape) == 2
    assert pred_mean.shape[0] == pred_std.shape[0]

    with np.errstate(divide='ignore'):
        val_z = (np.min(Y_train) - pred_mean) / (pred_std + jitter)
    val_ei = (np.min(Y_train) - pred_mean) * scipy.stats.norm.cdf(val_z) \
        + pred_std * scipy.stats.norm.pdf(val_z)
    val_aei = val_ei * (1.0 - noise / np.sqrt(pred_std**2 + noise**2))
    return val_aei

@utils_common.validate_types
def pure_exploit(pred_mean: np.ndarray) -> np.ndarray:
    """
    It is a pure exploitation criterion.

    :param pred_mean: posterior predictive mean function over `X_test`. Shape: (l, ).
    :type pred_mean: numpy.ndarray

    :returns: acquisition function values. Shape: (l, ).
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(pred_mean, np.ndarray)
    assert len(pred_mean.shape) == 1

    return -pred_mean

@utils_common.validate_types
def pure_explore(pred_std: np.ndarray) -> np.ndarray:
    """
    It is a pure exploration criterion.

    :param pred_std: posterior predictive standard deviation function over `X_test`. Shape: (l, ).
    :type pred_std: numpy.ndarray

    :returns: acquisition function values. Shape: (l, ).
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(pred_std, np.ndarray)
    assert len(pred_std.shape) == 1

    return pred_std



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: October 13, 2021
#
"""These files are written to implement tree-based regression models."""

from bayeso.trees import trees_generic_trees
from bayeso.trees import trees_random_forest


get_generic_trees = trees_generic_trees.get_generic_trees
get_random_forest = trees_random_forest.get_random_forest



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: October 8, 2021
#
"""It defines an abstract class of Bayesian optimization."""

import numpy as np
import abc
import time
import qmcpy

from bayeso import constants
from bayeso.utils import utils_bo
from bayeso.utils import utils_common
from bayeso.utils import utils_logger


class BaseBO(abc.ABC):
    """
    It is a Bayesian optimization class.

    :param range_X: a search space. Shape: (d, 2).
    :type range_X: numpy.ndarray
    :param str_surrogate: the name of surrogate model.
    :type str_surrogate: str.
    :param str_acq: the name of acquisition function.
    :type str_acq: str.
    :param str_optimizer_method_bo: the name of optimization method for
        Bayesian optimization.
    :type str_optimizer_method_bo: str.
    :param normalize_Y: flag for normalizing outputs.
    :type normalize_Y: bool.
    :param debug: flag for printing log messages.
    :type debug: bool.

    """

    def __init__(self,
        range_X: np.ndarray,
        str_surrogate: str,
        str_acq: str,
        str_optimizer_method_bo: str,
        normalize_Y: bool,
        debug: bool
    ):
        """
        Constructor method

        """

        assert isinstance(range_X, np.ndarray)
        assert isinstance(str_surrogate, str)
        assert isinstance(str_acq, str)
        assert isinstance(str_optimizer_method_bo, str)
        assert isinstance(normalize_Y, bool)
        assert isinstance(debug, bool)
        assert len(range_X.shape) == 2
        assert range_X.shape[1] == 2
        assert (range_X[:, 0] <= range_X[:, 1]).all()
        assert str_surrogate in constants.ALLOWED_SURROGATE \
            + constants.ALLOWED_SURROGATE_TREES
        assert str_acq in constants.ALLOWED_BO_ACQ
        assert str_optimizer_method_bo in constants.ALLOWED_OPTIMIZER_METHOD_BO \
            + constants.ALLOWED_OPTIMIZER_METHOD_BO_TREES

        self.range_X = range_X
        self.num_dim = range_X.shape[0]
        self.str_surrogate = str_surrogate
        self.str_acq = str_acq
        self.str_optimizer_method_bo = utils_bo.check_optimizer_method_bo(
            str_optimizer_method_bo, range_X.shape[0], debug)
        self.normalize_Y = normalize_Y
        self.logger = utils_logger.get_logger('bo_w_{}'.format(str_surrogate))
        self.debug = debug

    def _get_random_state(self, seed: constants.TYPING_UNION_INT_NONE):
        """
        It returns a random state, defined by `seed`.

        :param seed: None, or a random seed.
        :type seed: NoneType or int.

        :returns: a random state.
        :rtype: numpy.random.RandomState

        :raises: AssertionError

        """

        assert isinstance(seed, (int, constants.TYPE_NONE))

        if seed is not None:
            state_random = np.random.RandomState(seed)
        else:
            state_random = np.random.RandomState()

        return state_random

    def _get_bounds(self) -> constants.TYPING_LIST:
        """
        It returns list of range tuples, obtained from `self.range_X`.

        :returns: list of range tuples.
        :rtype: list

        """

        list_bounds = []
        for elem in self.range_X:
            list_bounds.append(tuple(elem))
        return list_bounds

    def _get_samples_grid(self, num_grids: int=constants.NUM_GRIDS_AO) -> np.ndarray:
        """
        It returns grids of `self.range_X`.

        :param num_grids: the number of grids.
        :type num_grids: int., optional

        :returns: grids of `self.range_X`. Shape: (`num_grids`:math:`^{\\text{d}}`, d).
        :rtype: numpy.ndarray

        :raises: AssertionError

        """

        assert isinstance(num_grids, int)

        initials = utils_common.get_grids(self.range_X, num_grids)
        return initials

    def _get_samples_uniform(self, num_samples: int,
        seed: constants.TYPING_UNION_INT_NONE=None
    ) -> np.ndarray:
        """
        It returns `num_samples` examples uniformly sampled.

        :param num_samples: the number of samples.
        :type num_samples: int.
        :param seed: None, or random seed.
        :type seed: NoneType or int., optional

        :returns: random examples. Shape: (`num_samples`, d).
        :rtype: numpy.ndarray

        :raises: AssertionError

        """

        assert isinstance(num_samples, int)
        assert isinstance(seed, (int, constants.TYPE_NONE))

        state_random = self._get_random_state(seed)

        list_initials = []
        for _ in range(0, num_samples):
            list_initial = []
            for elem in self.range_X:
                list_initial.append(state_random.uniform(elem[0], elem[1]))
            list_initials.append(np.array(list_initial))
        initials = np.array(list_initials)
        return initials

    def _get_samples_gaussian(self, num_samples: int,
        seed: constants.TYPING_UNION_INT_NONE=None
    ) -> np.ndarray:
        """
        It returns `num_samples` examples sampled from Gaussian distribution.

        :param num_samples: the number of samples.
        :type num_samples: int.
        :param seed: None, or random seed.
        :type seed: NoneType or int., optional

        :returns: random examples. Shape: (`num_samples`, d).
        :rtype: numpy.ndarray

        :raises: AssertionError

        """

        assert isinstance(num_samples, int)
        assert isinstance(seed, (int, constants.TYPE_NONE))

        state_random = self._get_random_state(seed)

        list_initials = []

        for _ in range(0, num_samples):
            list_initial = []
            for elem in self.range_X:
                new_mean = (elem[1] + elem[0]) / 2.0
                new_std = (elem[1] - elem[0]) / 4.0

                cur_sample = state_random.randn() * new_std + new_mean
                cur_sample = np.clip(cur_sample, elem[0], elem[1])

                list_initial.append(cur_sample)
            list_initials.append(np.array(list_initial))
        initials = np.array(list_initials)
        return initials

    def _get_samples_sobol(self, num_samples: int,
        seed: constants.TYPING_UNION_INT_NONE=None
    ) -> np.ndarray:
        """
        It returns `num_samples` examples sampled from Sobol' sequence.

        :param num_samples: the number of samples.
        :type num_samples: int.
        :param seed: None, or random seed.
        :type seed: NoneType or int., optional

        :returns: examples sampled from Sobol' sequence. Shape: (`num_samples`, d).
        :rtype: numpy.ndarray

        :raises: AssertionError

        """

        assert isinstance(num_samples, int)
        assert isinstance(seed, (int, constants.TYPE_NONE))

        sampler = qmcpy.Sobol(self.num_dim, seed=seed, graycode=True)
        samples = sampler.gen_samples(num_samples)

        samples = samples * (self.range_X[:, 1].flatten() - self.range_X[:, 0].flatten()) \
            + self.range_X[:, 0].flatten()
        return samples

    def _get_samples_halton(self, num_samples: int,
        seed: constants.TYPING_UNION_INT_NONE=None
    ) -> np.ndarray:
        """
        It returns `num_samples` examples sampled by Halton algorithm.

        :param num_samples: the number of samples.
        :type num_samples: int.
        :param seed: None, or random seed.
        :type seed: NoneType or int., optional

        :returns: examples sampled by Halton algorithm. Shape: (`num_samples`, d).
        :rtype: numpy.ndarray

        :raises: AssertionError

        """

        assert isinstance(num_samples, int)
        assert isinstance(seed, (int, constants.TYPE_NONE))

        sampler = qmcpy.Halton(self.num_dim, randomize='OWEN', generalize=False, seed=seed)
        samples = sampler.gen_samples(num_samples)

        samples = samples * (self.range_X[:, 1].flatten() - self.range_X[:, 0].flatten()) \
            + self.range_X[:, 0].flatten()
        return samples

    def get_samples(self, str_sampling_method: str,
        fun_objective: constants.TYPING_UNION_CALLABLE_NONE=None,
        num_samples: int=constants.NUM_SAMPLES_AO,
        seed: constants.TYPING_UNION_INT_NONE=None,
    ) -> np.ndarray:
        """
        It returns `num_samples` examples, sampled by a sampling method `str_sampling_method`.

        :param str_sampling_method: the name of sampling method.
        :type str_sampling_method: str.
        :param fun_objective: None, or objective function.
        :type fun_objective: NoneType or callable, optional
        :param num_samples: the number of samples.
        :type num_samples: int., optional
        :param seed: None, or random seed.
        :type seed: NoneType or int., optional

        :returns: sampled examples. Shape: (`num_samples`, d).
        :rtype: numpy.ndarray

        :raises: AssertionError

        """

        assert isinstance(str_sampling_method, str)
        assert callable(fun_objective) or fun_objective is None
        assert isinstance(num_samples, int)
        assert isinstance(seed, (int, constants.TYPE_NONE))
        assert str_sampling_method in constants.ALLOWED_SAMPLING_METHOD

        if str_sampling_method == 'grid':
            assert fun_objective is not None
            if self.debug:
                self.logger.debug('For this option, num_samples is used as num_grids.')
            samples = self._get_samples_grid(num_grids=num_samples)
            samples = utils_bo.get_best_acquisition_by_evaluation(samples, fun_objective)
        elif str_sampling_method == 'uniform':
            samples = self._get_samples_uniform(num_samples, seed=seed)
        elif str_sampling_method == 'gaussian':
            samples = self._get_samples_gaussian(num_samples, seed=seed)
        elif str_sampling_method == 'sobol':
            samples = self._get_samples_sobol(num_samples, seed=seed)
        elif str_sampling_method == 'halton':
            samples = self._get_samples_halton(num_samples, seed=seed)
        else:
            raise NotImplementedError('get_samples: allowed str_sampling_method,\
                but it is not implemented.')

        if self.debug:
            self.logger.debug('samples:\n%s', utils_logger.get_str_array(samples))

        return samples

    def get_initials(self, str_initial_method: str, num_initials: int,
        seed: constants.TYPING_UNION_INT_NONE=None,
    ) -> np.ndarray:
        """
        It returns `num_initials` examples, sampled by a sampling method `str_initial_method`.

        :param str_initial_method: the name of sampling method.
        :type str_initial_method: str.
        :param num_initials: the number of samples.
        :type num_initials: int.
        :param seed: None, or random seed.
        :type seed: NoneType or int., optional

        :returns: sampled examples. Shape: (`num_samples`, d).
        :rtype: numpy.ndarray

        :raises: AssertionError

        """

        assert isinstance(str_initial_method, str)
        assert isinstance(num_initials, int)
        assert isinstance(seed, (int, constants.TYPE_NONE))
        assert str_initial_method in constants.ALLOWED_INITIALIZING_METHOD_BO

        return self.get_samples(str_initial_method, num_samples=num_initials, seed=seed)

    @abc.abstractmethod
    def compute_posteriors(self): # pragma: no cover
        """
        It is an abstract method.

        """

        pass

    @abc.abstractmethod
    def compute_acquisitions(self): # pragma: no cover
        """
        It is an abstract method.

        """

        pass

    @abc.abstractmethod
    def optimize(self): # pragma: no cover
        """
        It is an abstract method.

        """

        pass



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: August 6, 2021
#
"""It defines a random forest."""

import numpy as np

from bayeso.trees import trees_common
from bayeso.utils import utils_common


@utils_common.validate_types
def get_random_forest(
    X: np.ndarray, Y: np.ndarray,
    num_trees: int,
    depth_max: int,
    size_min_leaf: int,
    num_features: int,
) -> list:
    """
    It returns a random forest.

    :param X: inputs. Shape: (N, d).
    :type X: np.ndarray
    :param Y: outputs. Shape: (N, 1).
    :type Y: str.
    :param num_trees: the number of trees.
    :type num_trees: int.
    :param depth_max: maximum depth of tree.
    :type depth_max: int.
    :param size_min_leaf: minimum size of leaf.
    :type size_min_leaf: int.
    :param num_features: the number of split features.
    :type num_features: int.

    :returns: list of trees
    :rtype: list

    :raises: AssertionError

    """

    assert isinstance(X, np.ndarray)
    assert isinstance(Y, np.ndarray)
    assert isinstance(num_trees, int)
    assert isinstance(depth_max, int)
    assert isinstance(size_min_leaf, int)
    assert isinstance(num_features, int)

    assert len(X.shape) == 2
    assert len(Y.shape) == 2
    assert X.shape[0] == Y.shape[0]
    assert Y.shape[1] == 1

    ratio_sample = 1.0
    replace_samples = True
    split_random_location = False

    list_trees = []

    for ind in range(0, num_trees):
        X_, Y_ = trees_common.subsample(X, Y, ratio_sample, replace_samples)

        root = trees_common._split(X_, Y_, num_features, split_random_location)
        trees_common.split(root, depth_max, size_min_leaf, num_features, split_random_location, 1)

        list_trees.append(root)

    return list_trees



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: October 13, 2021
#
"""It defines a common function for tree-based surrogates."""

import numpy as np
import multiprocessing
import itertools

from bayeso import constants
from bayeso.utils import utils_common


@utils_common.validate_types
def get_inputs_from_leaf(leaf: list) -> np.ndarray:
    """
    It returns an input from a leaf.

    :param leaf: pairs of input and output in a leaf.
    :type leaf: list

    :returns: an input. Shape: (n, d).
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(leaf, list)

    X = [bx for bx, by in leaf]
    return np.array(X)

@utils_common.validate_types
def get_outputs_from_leaf(leaf: list) -> np.ndarray:
    """
    It returns an output from a leaf.

    :param leaf: pairs of input and output in a leaf.
    :type leaf: list

    :returns: an output. Shape: (n, 1).
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(leaf, list)

    Y = [by for bx, by in leaf]
    return np.array(Y)

@utils_common.validate_types
def _mse(Y: np.ndarray) -> float:
    """
    It returns a mean squared loss over `Y`.

    :param Y: outputs in a leaf.
    :type Y: numpy.ndarray

    :returns: a loss value.
    :rtype: float

    :raises: AssertionError

    """

    assert isinstance(Y, np.ndarray)
    if len(Y.shape) == 2:
        assert Y.shape[1] == 1

    if Y.shape[0] > 0:
        mean = np.mean(Y, axis=0)
        mse_ = np.mean((Y - mean)**2)
    else:
        mse_ = 1e8

    return mse_

@utils_common.validate_types
def mse(left_right: tuple) -> float:
    """
    It returns a mean squared loss over `left_right`.

    :param left_right: a tuple of left and right leaves.
    :type left_right: tuple

    :returns: a loss value.
    :rtype: float

    :raises: AssertionError

    """

    assert isinstance(left_right, tuple)

    left, right = left_right

    Y_left = get_outputs_from_leaf(left)
    Y_right = get_outputs_from_leaf(right)

    mse_left = _mse(Y_left)
    mse_right = _mse(Y_right)

    return mse_left + mse_right

@utils_common.validate_types
def subsample(
    X: np.ndarray, Y: np.ndarray,
    ratio_sampling: float, replace_samples: bool
) -> constants.TYPING_TUPLE_TWO_ARRAYS:
    """
    It subsamples a bootstrap sample.

    :param X: inputs. Shape: (n, d).
    :type X: numpy.ndarray
    :param Y: outputs. Shape: (n, 1).
    :type Y: numpy.ndarray
    :param ratio_sampling: ratio of sampling.
    :type ratio_sampling: float
    :param replace_samples: a flag for sampling with replacement or without replacement.
    :type replace_samples: bool.

    :returns: a tuple of bootstrap sample. Shape: ((m, d), (m, 1)).
    :rtype: (numpy.ndarray, numpy.ndarray)

    :raises: AssertionError

    """

    assert isinstance(X, np.ndarray)
    assert isinstance(Y, np.ndarray)
    assert isinstance(ratio_sampling, float)
    assert isinstance(replace_samples, bool)

    assert len(X.shape) == 2
    assert len(Y.shape) == 2
    assert X.shape[0] == Y.shape[0]
    assert Y.shape[1] == 1

    if replace_samples:
        assert ratio_sampling > 0.0
    else:
        assert ratio_sampling > 0.0 and ratio_sampling <= 1.0

    num_X = X.shape[0]
    num_samples = int(num_X * ratio_sampling)

    indices = np.random.choice(num_X, num_samples, replace=replace_samples)

    X_ = X[indices]
    Y_ = Y[indices]

    return X_, Y_

@utils_common.validate_types
def _split_left_right(
    X: np.ndarray, Y: np.ndarray,
    dim_to_split: int, val_to_split: float
) -> tuple:
    """
    It splits `X` and `Y` to left and right leaves.

    :param X: inputs. Shape: (n, d).
    :type X: numpy.ndarray
    :param Y: outputs. Shape: (n, 1).
    :type Y: numpy.ndarray
    :param dim_to_split: a dimension to split.
    :type dim_to_split: int.
    :param val_to_split: a value to split.
    :type val_to_split: float

    :returns: a tuple of left and right leaves.
    :rtype: tuple

    :raises: AssertionError

    """

    assert isinstance(X, np.ndarray)
    assert isinstance(Y, np.ndarray)
    assert isinstance(dim_to_split, int)
    assert isinstance(val_to_split, (float, np.float16, np.float32, np.float64))

    assert len(X.shape) == 2
    assert len(Y.shape) == 2
    assert X.shape[0] == Y.shape[0]
    assert Y.shape[1] == 1

    left = []
    right = []

    for bx, by in zip(X, Y):
        if bx[dim_to_split] < val_to_split:
            left.append((bx, by))
        else:
            right.append((bx, by))
    return left, right

@utils_common.validate_types
def _split(
    X: np.ndarray, Y: np.ndarray,
    num_features: int, split_random_location: bool
) -> dict:
    """
    It splits `X` and `Y` to left and right leaves as a dictionary
    including split dimension and split location.

    :param X: inputs. Shape: (n, d).
    :type X: numpy.ndarray
    :param Y: outputs. Shape: (n, 1).
    :type Y: numpy.ndarray
    :param num_features: the number of features to split.
    :type num_features: int.
    :param split_random_location: flag for setting a split location randomly or not.
    :type split_random_location: bool.

    :returns: a dictionary of left and right leaves, spilt dimension, and split location.
    :rtype: dict.

    :raises: AssertionError

    """

    assert isinstance(X, np.ndarray)
    assert isinstance(Y, np.ndarray)
    assert isinstance(num_features, int)
    assert isinstance(split_random_location, bool)

    assert len(X.shape) == 2
    assert len(Y.shape) == 2
    assert X.shape[0] == Y.shape[0]
    assert X.shape[0] > 0
    assert Y.shape[1] == 1

    cur_ind = np.inf
    cur_val = np.inf
    cur_score = np.inf
    cur_left_right = None

    features = np.random.choice(X.shape[1], num_features, replace=False)

    for ind in features:
        dim_to_split = int(ind)
        num_evaluations = 1
        candidates_loc = np.sort(np.unique(X[:, dim_to_split]))

        if candidates_loc.shape[0] > 1:
            if split_random_location:
                min_bx = np.min(X[:, dim_to_split])
                max_bx = np.max(X[:, dim_to_split])
            else:
                num_evaluations = candidates_loc.shape[0] - 1

        for ind_loc in range(0, num_evaluations):
            if candidates_loc.shape[0] > 1:
                if split_random_location:
                    val_to_split = np.random.uniform(low=min_bx, high=max_bx)
                else:
                    val_to_split = np.mean(candidates_loc[ind_loc:ind_loc+2])
            else:
                val_to_split = X[0, dim_to_split]

            left_right = _split_left_right(X, Y, dim_to_split, val_to_split)
            left, right = left_right
            score = mse(left_right)

            if score < cur_score:
                cur_ind = dim_to_split
                cur_val = val_to_split
                cur_score = score
                cur_left_right = left_right

    return {
        'index': cur_ind,
        'value': cur_val,
        'left_right': cur_left_right
    }

@utils_common.validate_types
def split(
    node: dict,
    depth_max: int,
    size_min_leaf: int,
    num_features: int,
    split_random_location: bool,
    cur_depth: int
) -> constants.TYPE_NONE:
    """
    It splits a root node to construct a tree.

    :param node: a root node.
    :type node: dict.
    :param depth_max: maximum depth of tree.
    :type depth_max: int.
    :param size_min_leaf: minimum size of leaf.
    :type size_min_leaf: int.
    :param num_features: the number of split features.
    :type num_features: int.
    :param split_random_location: flag for setting a split location randomly or not.
    :type split_random_location: bool.
    :param cur_depth: depth of the current node.
    :type cur_depth: int.

    :returns: None.
    :rtype: NoneType

    :raises: AssertionError

    """

    assert isinstance(node, dict)
    assert isinstance(depth_max, int)
    assert isinstance(size_min_leaf, int)
    assert isinstance(num_features, int)
    assert isinstance(split_random_location, bool)
    assert isinstance(cur_depth, int)

    assert cur_depth > 0

    left, right = node['left_right']
    del(node['left_right'])

    if not left or not right: # pragma: no cover
        node['left'] = node['right'] = left + right
        return

    if cur_depth >= depth_max:
        node['left'], node['right'] = left, right
        return

    ##
    if len(left) <= size_min_leaf:
        node['left'] = left
    else:
        X_left = get_inputs_from_leaf(left)
        Y_left = get_outputs_from_leaf(left)

        node['left'] = _split(X_left, Y_left, num_features, split_random_location)
        split(node['left'], depth_max, size_min_leaf, num_features, split_random_location, cur_depth + 1)

    ##
    if len(right) <= size_min_leaf:
        node['right'] = right
    else:
        X_right = get_inputs_from_leaf(right)
        Y_right = get_outputs_from_leaf(right)

        node['right'] = _split(X_right, Y_right, num_features, split_random_location)
        split(node['right'], depth_max, size_min_leaf, num_features, split_random_location, cur_depth + 1)

@utils_common.validate_types
def _predict_by_tree(bx: np.ndarray, tree: dict) -> constants.TYPING_TUPLE_TWO_FLOATS:
    """
    It predicts a posterior distribution over `bx`, given `tree`.

    :param bx: an input. Shape: (d, ).
    :type bx: numpy.ndarray
    :param tree: a decision tree.
    :type tree: dict.

    :returns: posterior mean and standard devitation estimates.
    :rtype: (float, float)

    :raises: AssertionError

    """

    assert isinstance(bx, np.ndarray)
    assert isinstance(tree, dict)

    assert len(bx.shape) == 1

    if bx[tree['index']] < tree['value']:
        if isinstance(tree['left'], dict):
            return _predict_by_tree(bx, tree['left'])
        else:
            cur_Y = get_outputs_from_leaf(tree['left'])
            return np.mean(cur_Y), np.std(cur_Y)
    else:
        if isinstance(tree['right'], dict):
            return _predict_by_tree(bx, tree['right'])
        else:
            cur_Y = get_outputs_from_leaf(tree['right'])
            return np.mean(cur_Y), np.std(cur_Y)

@utils_common.validate_types
def _predict_by_trees(bx: np.ndarray, list_trees: list) -> constants.TYPING_TUPLE_TWO_FLOATS:
    """
    It predicts a posterior distribution over `bx`, given `list_trees`.

    :param bx: an input. Shape: (d, ).
    :type bx: numpy.ndarray
    :param list_trees: a list of decision trees.
    :type list_trees: list

    :returns: posterior mean and standard devitation estimates.
    :rtype: (float, float)

    :raises: AssertionError

    """

    assert isinstance(bx, np.ndarray)
    assert isinstance(list_trees, list)

    assert len(bx.shape) == 1

    list_mu_leaf = []
    list_sigma_leaf = []

    for tree in list_trees:
        mu_leaf, sigma_leaf = _predict_by_tree(bx, tree)

        list_mu_leaf.append(mu_leaf)
        list_sigma_leaf.append(sigma_leaf)

    mu = np.mean(list_mu_leaf)
    sigma = compute_sigma(np.array(list_mu_leaf), np.array(list_sigma_leaf))

    return mu, sigma

@utils_common.validate_types
def unit_predict_by_trees(X: np.ndarray, list_trees: list) -> constants.TYPING_TUPLE_TWO_ARRAYS:
    """
    It predicts a posterior distribution over `X`, given `list_trees`.

    :param X: inputs. Shape: (n, d).
    :type X: numpy.ndarray
    :param list_trees: a list of decision trees.
    :type list_trees: list

    :returns: posterior mean and standard devitation estimates. Shape: ((n, 1), (n, 1)).
    :rtype: (numpy.ndarray, numpy.ndarray)

    :raises: AssertionError

    """

    assert isinstance(X, np.ndarray)
    assert isinstance(list_trees, list)
    assert len(X.shape) == 2

    preds_mu = []
    preds_sigma = []

    for bx in X:
        mu_bx, sigma_bx = _predict_by_trees(bx, list_trees)

        preds_mu.append(mu_bx)
        preds_sigma.append(sigma_bx)

    preds_mu = np.array(preds_mu)[..., np.newaxis]
    preds_sigma = np.array(preds_sigma)[..., np.newaxis]

    return preds_mu, preds_sigma

@utils_common.validate_types
def predict_by_trees(X: np.ndarray, list_trees: list) -> constants.TYPING_TUPLE_TWO_ARRAYS:
    """
    It predicts a posterior distribution over `X`, given `list_trees`,
    using `multiprocessing`.

    :param X: inputs. Shape: (n, d).
    :type X: numpy.ndarray
    :param list_trees: a list of decision trees.
    :type list_trees: list

    :returns: posterior mean and standard devitation estimates. Shape: ((n, 1), (n, 1)).
    :rtype: (numpy.ndarray, numpy.ndarray)

    :raises: AssertionError

    """

    assert isinstance(X, np.ndarray)
    assert isinstance(list_trees, list)
    assert len(X.shape) == 2

    num_data_per_split = constants.NUM_DATA_PER_SPLIT_TREES
    num_cpu = multiprocessing.cpu_count()

    num_data = X.shape[0]

    if num_data <= num_data_per_split:
        preds_mu, preds_sigma = unit_predict_by_trees(X, list_trees)
    else:
        list_Xs = np.array_split(X, int(np.ceil(num_data / num_data_per_split)))

        with multiprocessing.Pool(num_cpu) as p:
            results = p.starmap(unit_predict_by_trees, zip(list_Xs, itertools.repeat(list_trees)))

        list_preds_mu, list_preds_sigma = zip(*results)

        preds_mu = np.concatenate(list_preds_mu, axis=0)
        preds_sigma = np.concatenate(list_preds_sigma, axis=0)

    return preds_mu, preds_sigma

@utils_common.validate_types
def compute_sigma(
    preds_mu_leaf: np.ndarray,
    preds_sigma_leaf: np.ndarray,
    min_sigma: float=0.0
) -> np.ndarray:
    """
    It computes predictive standard deviation estimates.

    :param preds_mu_leaf: predictive mean estimates of leaf. Shape: (n, ).
    :type preds_mu_leaf: numpy.ndarray
    :param preds_sigma_leaf: predictive standard deviation estimates of leaf. Shape: (n, ).
    :type preds_sigma_leaf: numpy.ndarray
    :param min_sigma: threshold for minimum standard deviation.
    :type min_sigma: float

    :returns: predictive standard deviation estimates. Shape: (n, ).
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(preds_mu_leaf, np.ndarray)
    assert isinstance(preds_sigma_leaf, np.ndarray)
    assert isinstance(min_sigma, float)

    assert len(preds_mu_leaf.shape) == 1
    assert len(preds_sigma_leaf.shape) == 1
    assert preds_mu_leaf.shape[0] == preds_sigma_leaf.shape[0]

    preds_sigma_leaf_ = np.maximum(preds_sigma_leaf, np.zeros(preds_sigma_leaf.shape) + min_sigma)

    sigma = np.mean(preds_mu_leaf**2 + preds_sigma_leaf_**2)
    sigma -= np.mean(preds_mu_leaf)**2

    if sigma < 0.0: # pragma: no cover
        sigma = 0.0
    sigma = np.sqrt(sigma)

    return sigma



from setuptools import setup
import os
import sys

if (sys.version_info.major == 3 and sys.version_info.minor == 6) or\
    (sys.version_info.major == 3 and sys.version_info.minor == 7) or\
    (sys.version_info.major == 3 and sys.version_info.minor == 8) or\
    (sys.version_info.major == 3 and sys.version_info.minor == 9):
    print('[SETUP] bayeso supports Python {}.{} version in this system.'.format(sys.version_info.major, sys.version_info.minor))
else:
    sys.exit('[ERROR] bayeso does not support Python {}.{} version in this system.'.format(sys.version_info.major, sys.version_info.minor))

path_requirements = 'requirements.txt'
list_packages = [
    'bayeso',
    'bayeso.bo',
    'bayeso.gp',
    'bayeso.tp',
    'bayeso.trees',
    'bayeso.wrappers',
    'bayeso.utils'
]

with open(path_requirements) as f:
    required = f.read().splitlines()

setup(
    name='bayeso',
    version='0.5.1',
    author='Jungtaek Kim',
    author_email='jtkim@postech.ac.kr',
    url='https://bayeso.org',
    license='MIT',
    description='Simple, but essential Bayesian optimization package',
    packages=list_packages,
    python_requires='>=3.6, <4',
    install_requires=required,
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ]
)



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: October 13, 2021
#
"""These files are for implementing Bayesian optimization classes.
"""

from bayeso.bo import bo_w_gp
from bayeso.bo import bo_w_tp
from bayeso.bo import bo_w_trees


BO = bo_w_gp.BOwGP
BOwGP = bo_w_gp.BOwGP
BOwTP = bo_w_tp.BOwTP
BOwTrees = bo_w_trees.BOwTrees



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: March 22, 2021
#
"""It defines Gaussian process regression."""

import time
import numpy as np
import scipy.stats

from bayeso import covariance
from bayeso import constants
from bayeso.gp import gp_kernel
from bayeso.utils import utils_gp
from bayeso.utils import utils_covariance
from bayeso.utils import utils_common
from bayeso.utils import utils_logger

logger = utils_logger.get_logger('gp')


@utils_common.validate_types
def sample_functions(mu: np.ndarray, Sigma: np.ndarray,
    num_samples: int=1
) -> np.ndarray:
    """
    It samples `num_samples` functions from multivariate Gaussian distribution (mu, Sigma).

    :param mu: mean vector. Shape: (n, ).
    :type mu: numpy.ndarray
    :param Sigma: covariance matrix. Shape: (n, n).
    :type Sigma: numpy.ndarray
    :param num_samples: the number of sampled functions
    :type num_samples: int., optional

    :returns: sampled functions. Shape: (num_samples, n).
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(mu, np.ndarray)
    assert isinstance(Sigma, np.ndarray)
    assert isinstance(num_samples, int)
    assert len(mu.shape) == 1
    assert len(Sigma.shape) == 2
    assert mu.shape[0] == Sigma.shape[0] == Sigma.shape[1]

    rv = scipy.stats.multivariate_normal(mean=mu, cov=Sigma)
    list_rvs = [rv.rvs() for _ in range(0, num_samples)]
    return np.array(list_rvs)

@utils_common.validate_types
def predict_with_cov(X_train: np.ndarray, Y_train: np.ndarray, X_test: np.ndarray,
    cov_X_X: np.ndarray, inv_cov_X_X: np.ndarray, hyps: dict,
    str_cov: str=constants.STR_COV,
    prior_mu: constants.TYPING_UNION_CALLABLE_NONE=None,
    debug: bool=False
) -> constants.TYPING_TUPLE_THREE_ARRAYS:
    """
    This function returns posterior mean and posterior standard deviation
    functions over `X_test`, computed by Gaussian process regression with
    `X_train`, `Y_train`, `cov_X_X`, `inv_cov_X_X`, and `hyps`.

    :param X_train: inputs. Shape: (n, d) or (n, m, d).
    :type X_train: numpy.ndarray
    :param Y_train: outputs. Shape: (n, 1).
    :type Y_train: numpy.ndarray
    :param X_test: inputs. Shape: (l, d) or (l, m, d).
    :type X_test: numpy.ndarray
    :param cov_X_X: kernel matrix over `X_train`. Shape: (n, n).
    :type cov_X_X: numpy.ndarray
    :param inv_cov_X_X: kernel matrix inverse over `X_train`. Shape: (n, n).
    :type inv_cov_X_X: numpy.ndarray
    :param hyps: dictionary of hyperparameters for Gaussian process.
    :type hyps: dict.
    :param str_cov: the name of covariance function.
    :type str_cov: str., optional
    :param prior_mu: None, or prior mean function.
    :type prior_mu: NoneType, or callable, optional
    :param debug: flag for printing log messages.
    :type debug: bool., optional

    :returns: a tuple of posterior mean function over `X_test`, posterior
        standard deviation function over `X_test`, and posterior covariance
        matrix over `X_test`. Shape: ((l, 1), (l, 1), (l, l)).
    :rtype: tuple of (numpy.ndarray, numpy.ndarray, numpy.ndarray)

    :raises: AssertionError

    """

    utils_gp.validate_common_args(X_train, Y_train, str_cov, prior_mu, debug, X_test)
    assert isinstance(cov_X_X, np.ndarray)
    assert isinstance(inv_cov_X_X, np.ndarray)
    assert isinstance(hyps, dict)
    assert len(cov_X_X.shape) == 2
    assert len(inv_cov_X_X.shape) == 2
    assert (np.array(cov_X_X.shape) == np.array(inv_cov_X_X.shape)).all()
    utils_covariance.check_str_cov('predict_with_cov', str_cov,
        X_train.shape, shape_X2=X_test.shape)

    prior_mu_train = utils_gp.get_prior_mu(prior_mu, X_train)
    prior_mu_test = utils_gp.get_prior_mu(prior_mu, X_test)
    cov_X_Xs = covariance.cov_main(str_cov, X_train, X_test, hyps, False)
    cov_Xs_Xs = covariance.cov_main(str_cov, X_test, X_test, hyps, True)
    cov_Xs_Xs = (cov_Xs_Xs + cov_Xs_Xs.T) / 2.0

    mu_Xs = np.dot(np.dot(cov_X_Xs.T, inv_cov_X_X), Y_train - prior_mu_train) + prior_mu_test
    Sigma_Xs = cov_Xs_Xs - np.dot(np.dot(cov_X_Xs.T, inv_cov_X_X), cov_X_Xs)
    return mu_Xs, np.expand_dims(np.sqrt(np.maximum(np.diag(Sigma_Xs), 0.0)), axis=1), Sigma_Xs

@utils_common.validate_types
def predict_with_hyps(X_train: np.ndarray, Y_train: np.ndarray, X_test: np.ndarray, hyps: dict,
    str_cov: str=constants.STR_COV,
    prior_mu: constants.TYPING_UNION_CALLABLE_NONE=None,
    debug: bool=False
) -> constants.TYPING_TUPLE_THREE_ARRAYS:
    """
    This function returns posterior mean and posterior standard deviation
    functions over `X_test`, computed by Gaussian process regression with
    `X_train`, `Y_train`, and `hyps`.

    :param X_train: inputs. Shape: (n, d) or (n, m, d).
    :type X_train: numpy.ndarray
    :param Y_train: outputs. Shape: (n, 1).
    :type Y_train: numpy.ndarray
    :param X_test: inputs. Shape: (l, d) or (l, m, d).
    :type X_test: numpy.ndarray
    :param hyps: dictionary of hyperparameters for Gaussian process.
    :type hyps: dict.
    :param str_cov: the name of covariance function.
    :type str_cov: str., optional
    :param prior_mu: None, or prior mean function.
    :type prior_mu: NoneType, or callable, optional
    :param debug: flag for printing log messages.
    :type debug: bool., optional

    :returns: a tuple of posterior mean function over `X_test`, posterior
        standard deviation function over `X_test`, and posterior covariance
        matrix over `X_test`. Shape: ((l, 1), (l, 1), (l, l)).
    :rtype: tuple of (numpy.ndarray, numpy.ndarray, numpy.ndarray)

    :raises: AssertionError

    """

    utils_gp.validate_common_args(X_train, Y_train, str_cov, prior_mu, debug, X_test)
    assert isinstance(hyps, dict)
    utils_covariance.check_str_cov('predict_with_hyps', str_cov,
        X_train.shape, shape_X2=X_test.shape)

    cov_X_X, inv_cov_X_X, _ = covariance.get_kernel_inverse(X_train,
        hyps, str_cov, debug=debug)
    mu_Xs, sigma_Xs, Sigma_Xs = predict_with_cov(X_train, Y_train, X_test,
        cov_X_X, inv_cov_X_X, hyps, str_cov=str_cov,
        prior_mu=prior_mu, debug=debug)

    return mu_Xs, sigma_Xs, Sigma_Xs

@utils_common.validate_types
def predict_with_optimized_hyps(X_train: np.ndarray, Y_train: np.ndarray, X_test: np.ndarray,
    str_cov: str=constants.STR_COV,
    str_optimizer_method: str=constants.STR_OPTIMIZER_METHOD_GP,
    prior_mu: constants.TYPING_UNION_CALLABLE_NONE=None,
    fix_noise: float=constants.FIX_GP_NOISE,
    debug: bool=False
) -> constants.TYPING_TUPLE_THREE_ARRAYS:
    """
    This function returns posterior mean and posterior standard deviation
    functions over `X_test`, computed by the Gaussian process regression
    optimized with `X_train` and `Y_train`.

    :param X_train: inputs. Shape: (n, d) or (n, m, d).
    :type X_train: numpy.ndarray
    :param Y_train: outputs. Shape: (n, 1).
    :type Y_train: numpy.ndarray
    :param X_test: inputs. Shape: (l, d) or (l, m, d).
    :type X_test: numpy.ndarray
    :param str_cov: the name of covariance function.
    :type str_cov: str., optional
    :param str_optimizer_method: the name of optimization method.
    :type str_optimizer_method: str., optional
    :param prior_mu: None, or prior mean function.
    :type prior_mu: NoneType, or callable, optional
    :param fix_noise: flag for fixing a noise.
    :type fix_noise: bool., optional
    :param debug: flag for printing log messages.
    :type debug: bool., optional

    :returns: a tuple of posterior mean function over `X_test`, posterior
        standard deviation function over `X_test`, and posterior covariance
        matrix over `X_test`. Shape: ((l, 1), (l, 1), (l, l)).
    :rtype: tuple of (numpy.ndarray, numpy.ndarray, numpy.ndarray)

    :raises: AssertionError

    """

    utils_gp.validate_common_args(X_train, Y_train, str_cov, prior_mu, debug, X_test)
    assert isinstance(str_optimizer_method, str)
    assert isinstance(fix_noise, bool)
    utils_covariance.check_str_cov('predict_with_optimized_kernel', str_cov,
        X_train.shape, shape_X2=X_test.shape)
    assert str_optimizer_method in constants.ALLOWED_OPTIMIZER_METHOD_GP

    time_start = time.time()

    cov_X_X, inv_cov_X_X, hyps = gp_kernel.get_optimized_kernel(X_train, Y_train,
        prior_mu, str_cov, str_optimizer_method=str_optimizer_method,
        fix_noise=fix_noise, debug=debug)
    mu_Xs, sigma_Xs, Sigma_Xs = predict_with_cov(X_train, Y_train, X_test,
        cov_X_X, inv_cov_X_X, hyps, str_cov=str_cov, prior_mu=prior_mu,
        debug=debug)

    time_end = time.time()
    if debug:
        logger.debug('time consumed to construct gpr: %.4f sec.', time_end - time_start)
    return mu_Xs, sigma_Xs, Sigma_Xs
